# methods in cngi_io.image.image should be used going forward,
# methods in cngi_io.image.cngi_image_io are deprecated
from .image import (
    load_image_block, make_empty_sky_image, read_image, write_image
)

__all__ = [
    'load_image_block', 'make_empty_sky_image', 'read_image', 'write_image'
]





from astropy import units as u
from astropy.io import fits
from astropy.time import Time
import dask.array as da
import logging
import numpy as np
import re
from ..common import __c

def __fits_image_to_xds_metadata(img_full_path:str, verbose:bool=False) -> dict:
    """
    TODO: complete documentation
    Create an xds without any pixel data from metadata from the specified FITS image
    """
    hdulist = fits.open(img_full_path)
    attrs, helpers, header = __fits_header_to_xds_attrs(hdulist)
    __create_coords(helpers, header)
    return attrs


def __fits_header_to_xds_attrs(hdulist: fits.hdu.hdulist.HDUList) -> dict:
    primary = None
    beams = None
    for hdu in hdulist:
        if hdu.name == 'PRIMARY':
            primary = hdu
        elif hdu.name == 'BEAMS':
            beams = hdu
        else:
            raise RuntimeError(f'Unknown HDU name {hdu.name}')
    logging.warn('** cb')
    if not primary:
        raise RuntimeError(f'No PRIMARY HDU found in fits file')
    dir_axes = None
    header = primary.header
    helpers = {}
    attrs = {}
    naxes = header.get('NAXIS')
    helpers['naxes'] = naxes
    # fits indexing starts at 1, not 0
    t_axes = np.array([0,0])
    dim_map = {}
    ctypes = []
    shape = []
    crval = []
    cdelt = []
    crpix = []
    cunit = []
    for i in range(1, naxes+1):
        ax_type = header[f'CTYPE{i}']
        if ax_type.startswith('RA-'):
            t_axes[0] = i
        elif ax_type.startswith('DEC-'):
            t_axes[1] = i
        elif ax_type == 'STOKES':
            dim_map['pol'] = i - 1
        elif ax_type.startswith('FREQ') or ax_type == 'VOPT':
            dim_map['freq'] = i - 1
        else:
            raise RuntimeError(f'{ax_type} is an unsupported axis')
        ctypes.append(ax_type)
        shape.append(header[f'NAXIS{i}'])
        crval.append(header[f'CRVAL{i}'])
        cdelt.append(header[f'CDELT{i}'])
        # FITS 1-based to python 0-based
        crpix.append(header[f'CRPIX{i}'] - 1)
        cunit.append(header[f'CUNIT{i}'])
    helpers['ctype'] = ctypes
    helpers['shape'] = shape
    helpers['crval'] = crval
    helpers['cdelt'] = cdelt
    helpers['crpix'] = crpix
    helpers['cunit'] = cunit
    if 'RESTFRQ' in header:
        helpers['restfreq'] = header['RESTFRQ']
    if (t_axes > 0).all():
        dir_axes = t_axes[:]
        dir_axes = dir_axes - 1
        helpers['dir_axes'] = dir_axes
        dim_map['l'] = dir_axes[0]
        dim_map['m'] = dir_axes[1]
        helpers['dim_map'] = dim_map
    else:
        raise RuntimeError('Could not find both direction axes')
    logging.warn('** cd')
    if dir_axes is not None:
        p0 = header.get(f'CTYPE{t_axes[0]}')[-3:]
        p1 = header.get(f'CTYPE{t_axes[1]}')[-3:]
        if p0 != p1:
            raise RuntimeError(
                f'Projections for direction axes ({p0}, {p1}) differ, but they '
                'must be the same'
            )
        direction = {}
        direction['projection'] = p0
        ref_sys = header['RADESYS']
        ref_eqx = header['EQUINOX']
        # fits does not support conversion frames
        direction['conversion_system'] = ref_sys
        direction['conversion_equinox'] = ref_eqx
        direction['system'] = ref_sys
        logging.warn('** ce')
        direction['equinox'] = ref_eqx
        logging.warn('** da')
        deg_to_rad = np.pi/180.0
        logging.warn('** db')
        direction['latpole'] = header.get('LATPOLE') * deg_to_rad
        direction['longpole'] = header.get('LONPOLE') * deg_to_rad
        pc = np.zeros([2,2])
        for i in (0, 1):
            for j in (0, 1):
                pc[i][j] = header.get(f'PC{dir_axes[i]}_{dir_axes[j]}')
        direction['pc'] = pc
        attrs['direction'] = direction
    # FIXME read fits data in chunks in case all data too large to hold in memory
    has_mask = da.any(da.isnan(primary.data)).compute()
    attrs['active_mask'] = 'mask0' if has_mask else None
    if 'BMAJ' in header.keys():
        # single global beam
        attrs['beam'] = {
            'bmaj': {'unit': 'arcsec', 'value': header.get('BMAJ')},
            'bmin': {'unit': 'arcsec', 'value': header.get('BMIN')},
            'positionangle': {'unit': 'arcsec', 'value': header.get('BPA')}
        }
    elif 'CASAMBM' in header.keys() and header.get('CASAMBM'):
        # multi-beam
        pass
    else:
        # no beam
        attrs['beam'] = None
    attrs['object'] = header.get('OBJECT') if 'OBJECT' in header else None
    obsdate = {}
    obsdate['value'] = Time(header.get('DATE-OBS'), format='isot').mjd
    obsdate['unit'] = 'd'
    obsdate['refer'] = 'UTC1'
    attrs['obsdate'] = obsdate
    helpers['obsdate'] = obsdate
    attrs['observer'] = header.get('OBSERVER')
    long_unit = header.get(f'CUNIT{t_axes[0]}')
    lat_unit = header.get(f'CUNIT{t_axes[1]}')
    unit = []
    for uu in [long_unit, lat_unit]:
        if uu == 'deg':
            unit.append(u.deg)
        elif uu == 'rad':
            unit.append(u.rad)
        else:
            raise RuntimeError(f'Unsupported direction unit {uu}')
    pc_long = float(header.get(f'CRVAL{t_axes[0]}')) * unit[0]
    pc_lat = float(header.get(f'CRVAL{t_axes[1]}')) * unit[1]
    pc_long = pc_long.to(u.rad).value
    pc_lat = pc_lat.to(u.rad).value
    attrs['pointing_center'] = {
        'value': np.array([pc_long, pc_lat]), 'initial': True
    }
    attrs['description'] = ''
    tel = {}
    tel['name'] = header.get('TELESCOP')
    x = header.get('OBSGEO-X')
    y = header.get('OBSGEO-Y')
    z = header.get('OBSGEO-Z')
    xyz = np.array([x, y, z])
    r = np.sqrt(np.sum(xyz*xyz))
    lat = np.arcsin(z/r)
    long = np.arctan2(y, x)
    tel['position'] = {
        'type': 'position', 'refer': 'ITRF',
        'm2': {'value': r, 'unit': 'm'},
        'm1': {'unit': 'rad', 'value': lat},
        'm0': {'unit': 'rad', 'value': long}
    }
    attrs['telescope'] = tel
    # TODO complete __make_history_xds when spec has been finalized
    # attrs['history'] = __make_history_xds(header)
    exclude = [
        'ALTRPIX', 'ALTRVAL', 'BITPIX', 'BSCALE', 'BTYPE', 'BUNIT',
        'BZERO', 'CASAMBM', 'DATE', 'DATE-OBS', 'EQUINOX', 'EXTEND',
        'HISTORY', 'LATPOLE', 'LONPOLE', 'OBSERVER', 'ORIGIN', 'TELESCOP',
        'OBJECT', 'RADESYS', 'RESTFRQ', 'SIMPLE', 'SPECSYS', 'TIMESYS',
        'VELREF'
    ]
    regex = r'|'.join([
        '^NAXIS\d?$', '^CRVAL\d$', '^CRPIX\d$', '^CTYPE\d$', '^CDELT\d$',
        '^CUNIT\d$', '^OBSGEO-(X|Y|Z)$', '^P(C|V)\d_\d$'
    ])
    user = {}
    for (k, v) in header.items():
        if re.search(regex, k) or k in exclude:
            continue
        user[k.lower()] = v
    attrs['user'] = user
    print(attrs)
    return attrs, helpers, header


def __make_history_xds(header):
    # TODO complete writing history when we actually have a spec for what
    # the image history is supposed to be, since doing this now may
    # be a waste of time if the final spec turns out to be significantly
    # different from our current ad hoc history xds
    # in astropy, 3506803168 seconds corresponds to 1970-01-01T00:00:00
    history_list = list(header.get('HISTORY'))
    for i in range(len(history_list)-1, -1, -1):
        if (
            (
                i == len(history_list) - 1
                and history_list[i] == 'CASA END LOGTABLE'
            ) or (
                i == 0
                and history_list[i] == 'CASA START LOGTABLE'
            )
        ):
            history_list.pop(i)
        elif history_list[i].startswith('>'):
            # entry continuation line
            history_list[i-1] = history_list[i-1] + history_list[i][1:]
            history_list.pop(i)


def __create_coords(helpers, header):
    dir_axes = helpers['dir_axes']
    dim_map = helpers['dim_map']
    sphr_dims = (
        [dim_map['l'], dim_map['m']]
        if ('l' in dim_map) and ('m' in dim_map)
        else []
    )
    print('sph_dims', sphr_dims)
    coords = {}
    coords['time'] = __get_time_values(helpers)
    coords['pol'] = __get_pol_values(helpers)
    coords['freq'] = __get_freq_values(helpers)
    coords['vel'] = (['freq'], __get_velocity_values(helpers))
    print('vel', coords['vel'][1])


def __get_time_values(helpers):
    return [ helpers['obsdate']['value'] ]


def __get_pol_values(helpers):
    # as mapped in casacore Stokes.h
    stokes_map = [
        'Undefined', 'I', 'Q', 'U', 'V',
        'RR', 'RL', 'LR', 'LL',
        'XX', 'XY', 'YX', 'YY',
        'RX', 'RY', 'LX', 'LY',
        'XR', 'XL', 'YR', 'YL',
        'PP', 'PQ'
    ]
    idx = helpers['ctype'].index('STOKES')
    if idx >= 0:
        vals = []
        crval = int(helpers['crval'][idx])
        crpix = int(helpers['crpix'][idx])
        cdelt = int(helpers['cdelt'][idx])
        stokes_start_idx = crval - cdelt*crpix
        for i in range(helpers['shape'][idx]):
            stokes_idx = (stokes_start_idx + i)*cdelt
            vals.append(stokes_map[stokes_idx])
        return vals
    else:
        return ['I']


def __get_freq_values(helpers:dict) -> list:
    print('** ia')
    vals = []
    ctype = helpers['ctype']
    print('** ib')
    if 'FREQ' in ctype:
        freq_idx = ctype.index('FREQ')
        crval = helpers['crval'][freq_idx]
        crpix = helpers['crpix'][freq_idx]
        cdelt = helpers['cdelt'][freq_idx]
        cunit = helpers['cunit'][freq_idx]
        freq_start_val = crval - cdelt*crpix
        for i in range(helpers['shape'][freq_idx]):
            vals.append(freq_start_val + i*cdelt)
        helpers['freq'] = vals * u.Unit(cunit)
        return vals
    elif 'VOPT' in ctype:
        if 'restfreq' in helpers:
            restfreq = helpers['restfreq'] * u.Hz
        else:
            raise RuntimeError(
                'Spectral axis in FITS header is velocity, but there is '
                'no rest frequency so converting to frequency is not possible'
            )
        helpers['doppler'] = 'optical'
        v_idx = ctype.index('VOPT')
        crval = helpers['crval'][v_idx]
        crpix = helpers['crpix'][v_idx]
        cdelt = helpers['cdelt'][v_idx]
        cunit = helpers['cunit'][v_idx]
        v_start_val = crval - cdelt*crpix
        vel = []
        for i in range(helpers['shape'][v_idx]):
            vel.append(v_start_val + i*cdelt)
        vel = vel * u.Unit(cunit)
        print('** ie')
        # (-1 + f0/f) = v/c
        # f0/f = v/c + 1
        # f = f0/(v/c + 1)
        print('** ja')
        freq = restfreq/(np.array(vel.value) * vel.unit/__c + 1)
        print('** if')
        freq = freq.to(u.Hz)
        helpers['vel'] = vel
        print('** ig')
        return list(freq.value)
    else:
        return [1420e6]


def __get_velocity_values(helpers:dict) -> list:
    if 'vel' in helpers:
        return helpers['vel'].to(u.m/u.s).value
    elif 'freq' in helpers:
        if helpers['doppler'] == 'optical':
            # (-1 + f0/f) = v/c
            v = (helpers['restfreq']/helpers['freq'].to('Hz').value - 1) * __c
            v = v.to(u.m)
            helpers['vel'] = v
            return v.value







import astropy
from astropy import units as u
from casacore import images, quanta, tables
from casacore.images import coordinates
import dask
import dask.array as da
import logging
import numpy as np
import os
from typing import Union
import xarray as xr

from .common import (
    __active_mask, __doppler_types, __native_types, __object_name,
    __pointing_center
)
from ..common import __dask_arrayize

__image_type = 'image_type'


def __add_coord_attrs(xds: xr.Dataset, icoords: dict, diraxes: list) -> xr.Dataset:
    xds = __add_time_attrs(xds, icoords)
    xds = __add_freq_attrs(xds, icoords)
    xds = __add_vel_attrs(xds, icoords)
    xds = __add_dir_lin_attrs(xds, icoords, diraxes)
    return xds


def __add_dir_lin_attrs(xds, coord_dict, dir_axes):
    for k in coord_dict:
        if k.startswith('direction'):
            dd = coord_dict[k]
            for i in (0, 1):
                meta = {}
                meta['unit'] = 'rad'
                unit = dd['units'][i]
                if unit == "'":
                    unit = 'arcmin'
                elif unit == '"':
                    unit = 'arcsec'
                ap_unit = 1 * u.Unit(unit)
                scale = ap_unit.to('rad').value
                meta['wcs'] = {}
                meta['wcs']['crval'] = dd['crval'][i]*scale
                meta['wcs']['cdelt'] = dd['cdelt'][i]*scale
                xds[dir_axes[i]].attrs = meta
            break
        elif k.startswith('linear'):
            ld = coord_dict[k]
            for i in (0, 1):
                meta = {}
                meta['unit'] = ld['units'][i]
                meta['wcs'] = {}
                meta['wcs']['crval'] = ld['crval'][i]
                meta['wcs']['cdelt'] = ld['cdelt'][i]
                xds[dir_axes[i]].attrs = meta
            break
    return xds


def __add_freq_attrs(xds, coord_dict):
    freq_coord = xds['freq']
    meta = {}
    for k in coord_dict:
        if k.startswith('spectral'):
            sd = coord_dict[k]
            meta['conversion'] = sd['conversion']
            meta['native_type'] = __native_types[sd['nativeType']]
            meta['restfreq'] = sd['restfreq']
            meta['restfreqs'] = sd['restfreqs']
            meta['system'] = sd['system']
            meta['unit'] = sd['unit']
            meta['wave_unit'] = sd['waveUnit']
            meta['wcs'] = {}
            meta['wcs']['crval'] = sd['wcs']['crval']
            meta['wcs']['cdelt'] = sd['wcs']['cdelt']
            break
    if not meta:
		# this is the default frequency information CASA creates
        meta = {
			'conversion': {
				'direction': {
					'm0': {'unit': 'rad', 'value': 0.0},
   					'm1': {'unit': 'rad', 'value': 1.5707963267948966},
   					'refer': 'J2000',
   					'type': 'direction'
				},
  				'epoch': {
					'm0': {'unit': 'd', 'value': 0.0},
   					'refer': 'LAST',
   					'type': 'epoch'
				},
  				'position': {
					'm0': {'unit': 'rad', 'value': 0.0},
   					'm1': {'unit': 'rad', 'value': 0.0},
   					'm2': {'unit': 'm', 'value': 0.0},
   					'refer': 'ITRF',
   					'type': 'position'
				},
  				'system': 'LSRK'
			},
 			'nativeType': 'FREQ',
 			'restfreq': 1420405751.7860003,
 			'restfreqs': [1420405751.7860003],
 			'system': 'LSRK',
 			'unit': 'Hz',
 			'waveUnit': 'mm',
 			'wcs': {
				'cdelt': 1000.0,
  				'crval': 1415000000.0,
			}
		}
    freq_coord.attrs = meta
    xds['freq'] = freq_coord
    return xds


def __add_mask(
    xds:xr.Dataset, name:str, ary:Union[np.ndarray, da.array], dimorder:list
) -> xr.Dataset:
    xda = xr.DataArray(ary, dims=dimorder)
    # True pixels are good in numpy masked arrays
    xda = da.logical_not(xda)
    xda.attrs['image_type'] = 'Mask'
    xda = xda.rename(name)
    xds[xda.name] = xda
    return xds


def __add_sky_or_apeture(
        xds: xr.Dataset, ary: Union[np.ndarray, da.array],
        dimorder:list, img_full_path: str, has_sph_dims:bool
) -> xr.Dataset:
    xda = xr.DataArray(ary, dims=dimorder)
    casa_image = images.image(img_full_path)
    image_type = casa_image.info()['imageinfo']['imagetype']
    unit = casa_image.unit()
    del casa_image
    xda.attrs[__image_type] = image_type
    xda.attrs['unit'] = unit
    name = 'sky' if has_sph_dims else 'apeture'
    xda = xda.rename(name)
    xds[xda.name] = xda
    return xds


def __get_time_format(value:float, unit:str) -> str:
    if value >= 40000 and value <= 100000 and unit == 'd':
        return  'MJD'
    else:
        return ''



def __add_time_attrs(xds: xr.Dataset, coord_dict: dict) -> xr.Dataset:
    time_coord = xds['time']
    meta = {}
    meta['refer'] = coord_dict['obsdate']['refer']
    meta['unit'] = coord_dict['obsdate']['m0']['unit']
    meta['format'] = __get_time_format(time_coord[0], meta['unit'])
    time_coord.attrs = meta
    xds['time'] = time_coord
    return xds


def __add_vel_attrs(xds, coord_dict):
    vel_coord = xds['vel']
    meta = {'unit': 'm/s'}
    for k in coord_dict:
        if k.startswith('spectral'):
            sd = coord_dict[k]
            meta['doppler_type'] = __doppler_types[sd['velType']]
            break
    if not meta:
            meta['doppler_type'] = __doppler_types[0]
    vel_coord.attrs = meta
    xds['vel'] = vel_coord
    return xds


def __casa_image_to_xds_attrs(img_full_path: str, history: bool=True) -> dict:
    """
    Get the xds level attributes as a python dictionary
    """
    casa_image = images.image(img_full_path)
    meta_dict = casa_image.info()
    del casa_image
    coord_dict = meta_dict['coordinates']
    attrs = {}
    dir_key = None
    for k in coord_dict.keys():
        if (k.startswith('direction')):
            dir_key = k
            break
    if dir_key:
        # shared direction coordinate attributes
        coord_dir_dict = coord_dict[dir_key]
        system = 'system'
        if system not in coord_dir_dict:
            raise Exception('No direction reference frame found')
        dir_dict = {}
        casa_system = coord_dir_dict[system]
        ap_system, ap_equinox = __convert_direction_system(casa_system, 'native')
        dir_dict[system] = ap_system
        dir_dict['equinox'] = ap_equinox if ap_equinox else None
        dir_dict['conversion_system'] = None

        cs = 'conversionSystem'
        if cs in coord_dir_dict and (coord_dir_dict[cs] != coord_dir_dict[system]):
                logging.warn(
                    'Conversion direction frame differs from native direction '
                    'frame in CASA image. However, ngCASA does not support conversion '
                    'frames at this time so the ngCASA image\'s conversion frame '
                    'will be set to the native frame'
                )
        dir_dict['conversion_system'] = dir_dict[system]
        dir_dict['conversion_equinox'] = dir_dict['equinox']
        k = 'latpole'
        if k in coord_dir_dict:
            deg_to_rad = np.pi/180
            for j in (k, 'longpole'):
                dir_dict[j] = {
                    'value': coord_dir_dict[j]*deg_to_rad, 'unit': 'rad'
                }
        for j in ('pc', 'projection_parameters', 'projection'):
            if j in coord_dir_dict:
                dir_dict[j] = coord_dir_dict[j]
        attrs['direction'] = dir_dict
    attrs['telescope'] = {}
    telescope = attrs['telescope']
    attrs['obsdate'] = {}
    obsdate = attrs['obsdate']
    attrs[__pointing_center] = coord_dict['pointingcenter'].copy()
    for k in ('observer', 'obsdate', 'telescope', 'telescopeposition'):
        if k.startswith('telescope'):
            if k == 'telescope':
                telescope['name'] = coord_dict[k]
            else:
                telescope['position'] = coord_dict[k]
        elif k == 'obsdate':
            obsdate['refer'] = coord_dict[k]['refer']
            obsdate['unit'] = coord_dict[k]['m0']['unit']
            obsdate['value'] = coord_dict[k]['m0']['value']
            obsdate['format'] = __get_time_format(obsdate['value'], obsdate['unit'])
        else:
            attrs[k] = coord_dict[k] if k in coord_dict else ''
    imageinfo = meta_dict['imageinfo']
    obj = 'objectname'
    attrs[__object_name] = imageinfo[obj] if obj in imageinfo else ''
    attrs['beam'] = __get_beam(imageinfo)
    attrs['user'] = meta_dict['miscinfo']
    casa_table = tables.table(
        img_full_path, readonly=True,
        lockoptions={'option': 'usernoread'}, ack=False
    )
    defmask = 'Image_defaultmask'
    attrs[__active_mask] = (
            casa_table.getkeyword(defmask)
            if defmask in casa_table.keywordnames()
            else None
    )
    casa_table.close()
    attrs['description'] = None
    # if also loading history, put it as another xds in the attrs
    if history:
        htable = os.sep.join([img_full_path, 'logtable'])
        if os.path.isdir(htable):
            attrs['history'] = read_generic_table(htable)
        else:
            logging.warn(
                f'Unable to find history table {htable}. '
                'History will not be included'
            )
    return attrs


def __casa_image_to_xds_metadata(img_full_path:str, verbose:bool=False) -> dict:
    """
    TODO: complete documentation
    Create an xds without any pixel data from metadata from the specified CASA image
    """
    attrs = {}
    casa_image = images.image(img_full_path)
    # shape list is the reverse of the actual image shape
    shape = casa_image.shape()[::-1]
    attrs['shape'] = shape
    meta_dict = casa_image.info()
    coord_dict = meta_dict['coordinates']
    axis_names = __flatten_list(casa_image.coordinates().get_axes())[::-1]
    attrs['icoords'] = coord_dict
    diraxes = [
        aa.lower().replace(' ', '_') for cc in coord_dict.items()
        if (cc[0][:-1] in ['direction', 'linear']) and len(cc[1]['axes']) >= 2
        for aa in cc[1]['axes']
    ]
    attrs['dir_axes'] = diraxes
    dimmap = __get_dimmap(coord_dict, verbose)
    attrs['dimmap'] = dimmap
    sphr_dims = (
        [dimmap['l'], dimmap['m']]
        if ('l' in dimmap) and ('m' in dimmap)
        else []
    )
    attrs['sphr_dims'] = sphr_dims
    coords = {}
    coords['time'] = __get_time_values(coord_dict)
    coords['pol'] = __get_pol_values(coord_dict)
    coords['freq'] = __get_freq_values(casa_image.coordinates(), shape)
    coords['vel'] = (['freq'], __get_velocity_values(coord_dict, coords['freq']))
    if len(sphr_dims) > 0:
        for k in coord_dict.keys():
            if k.startswith('direction'):
                dc = coordinates.directioncoordinate(coord_dict[k])
                break
        l_world, m_world = __compute_world_sph_dims(
            sphr_dims, diraxes, casa_image.coordinates(), dc, shape, dimmap
        )
        coords[l_world[0]] = (['l', 'm'], l_world[1])
        coords[m_world[0]] = (['l', 'm'], m_world[1])
    else:
        # Fourier image
        coords['u'], coords['v'] = __get_uv_values(coord_dict, axis_names, shape)
    del casa_image
    xds = xr.Dataset(coords=coords)
    attrs['xds'] = xds
    return attrs


def __compute_world_sph_dims(
    sphr_dims: list, dir_axes: list, csys:dict,
    dc:coordinates.directioncoordinate, shape:tuple, dimmap:dict
) -> list:
    proj = dc.get_projection()
    # casacore csys getters return values in opposite order as the real axes order
    coord_names = csys.get_names()[::-1]
    for i, name in enumerate(coord_names):
        if name.startswith('direction'):
            dc_index = i
            break
    unit = csys.get_unit()[::-1]
    inc = csys.get_increment()[::-1]
    ref_pix = csys.get_referencepixel()[::-1]
    ref_val = csys.get_referencevalue()[::-1]
    wcs_dict = {}
    # opposite of what you expect because, even though the coordinates have
    # been ordered coorectly, the two direction coordinate axes are still flipped
    for i, name in enumerate(dir_axes[::-1]):
        if name.startswith('right'):
            long_axis_name = name
            fi = 1
            wcs_dict[f'CTYPE1'] = f'RA---{proj}'
            wcs_dict[f'NAXIS1'] = shape[dimmap['l']]
        if name.startswith('dec'):
            lat_axis_name = name
            fi = 2
            wcs_dict['CTYPE2'] = f'DEC--{proj}'
            wcs_dict[f'NAXIS2'] = shape[dimmap['m']]
        t_unit = unit[dc_index][i]
        if t_unit == "'":
            t_unit = 'arcmin'
        elif t_unit == '"':
            t_unit = 'arcsec'
        wcs_dict[f'CUNIT{fi}'] = t_unit
        wcs_dict[f'CDELT{fi}'] = inc[dc_index][i]
        # FITS arrays are 1-based
        wcs_dict[f'CRPIX{fi}'] = ref_pix[dc_index][i] + 1
        wcs_dict[f'CRVAL{fi}'] = ref_val[dc_index][i]
    w = astropy.wcs.WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    long, lat = w.pixel_to_world_values(x, y)
    # long, lat from above eqn will always be in degrees, so convert to rad
    f = np.pi/180
    long *= f
    lat *= f
    return [[long_axis_name, long], [lat_axis_name, lat]]


def __convert_beam_to_rad(beam: dict) -> dict:
    """Convert a beam dictionary to radians"""
    mybeam = {}
    for k in beam:
        q = quanta.quantity(beam[k])
        q.convert(quanta.quantity('1rad'))
        j = 'pa' if k == 'positionangle' else k
        mybeam[j] = q.to_dict()
    return mybeam


def __convert_direction_system(casa_system: str, which: str) -> tuple:
    if casa_system == 'J2000':
        logging.info(
            f'J2000 found as {which} reference frame in CASA image '
            'This corresponds to FK5(equinox="J2000") in astropy. '
            'Metadata will be written appropriately'
        )
        return ('FK5', 'J2000')
    elif casa_system == 'B1950':
        logging.info(
            f'B1950 found as {which} reference frame in CASA image '
            'This corresponds to FK4(equinox="B1950") in astropy. '
            'Metadata will be written appropriately'
        )
        return ('FK4', 'B1950')
    elif casa_system in ('GALACTIC', 'ICRS'):
        return (casa_system, None)
    else:
        raise Exception(
            f'astropy does not support frame {casa_system} and this '
            'application does not support converting it to '
            'something astropy does support. You can try to regridding '
            'it to ICRS, GALACTIC, J2000, or B1950 in CASA and then '
            're-run this application on the regridded image'
        )


####################################
# local helper - return a dictionary of table attributes created from keywords and column descriptions
def extract_table_attributes(infile):
    tb_tool = tables.table(infile, readonly=True, lockoptions={'option': 'usernoread'}, ack=False)
    kwd = tb_tool.getkeywords()
    attrs = dict([(kk, kwd[kk]) for kk in kwd if kk not in os.listdir(infile)])
    cols = tb_tool.colnames()
    column_descriptions = {}
    for col in cols:
        column_descriptions[col] = tb_tool.getcoldesc(col)
    attrs['column_descriptions'] = column_descriptions
    attrs['info'] = tb_tool.info()
    tb_tool.close()
    return attrs


def __flatten_list(list_of_lists: list) -> list:
    flat = []
    for x in list_of_lists:
        if type(x) == list:
            flat.extend(__flatten_list(x))
        else:
            flat.append(x)
    return flat


def __get_beam(imageinfo: dict):
    """Returns None if no beam. Multiple beams are handled elsewhere"""
    k = 'restoringbeam'
    key = None
    if k in imageinfo and 'major' in imageinfo[k]:
        return __convert_beam_to_rad(imageinfo[k])
    return None


def __get_chunk_list(chunks:dict, coords:list, image_shape:Union[list, tuple]) -> tuple:
    ret_list = list(image_shape)
    axis = 0
    for c in coords:
        if c == 'direction' or c == 'linear':
            lm = ('l', 'm')
            uv = ('u', 'v')
            for i in (0, 1):
                for k in (lm[i], uv[i]):
                    if k in chunks:
                        ret_list[axis] = chunks[k]
                        break
                # add an axis because direction has 2 axes
                if i == 0:
                    axis += 1
        elif c == 'spectral':
            if 'freq' in chunks:
                ret_list[axis] = chunks['freq']
        elif c == 'stokes':
            if 'pol' in chunks:
                ret_list[axis] = chunks['pol']
        else:
            raise Exception(f'Unhandled coordinate type {c}')
        axis += 1
    return tuple(ret_list)


def __get_dimmap(coords: list, verbose: bool=False) -> dict:
    # example of dimmap:
    # [('direction0', 0), ('direction1', 1), ('spectral0', 2), ('stokes0', 3)]
    dimmap : list[tuple] = [
        (coord[:-1]+str(ii), ci) for coord in coords
        if coord[:-1] in ['direction', 'stokes', 'spectral', 'linear']
        for ii, ci in enumerate(coords['pixelmap%s' % coord[-1]])
    ]
    if verbose:
        print(f'dimmap: {dimmap}')
    # example of dimmap after next statment
    # [('l', 0), ('m', 1), ('chan', 2), ('pol', 3)]
    dimmap = [
        (rr[0].replace(
            'stokes0','pol'
        ).replace('spectral0','chan').replace('direction0','l').replace(
            'direction1','m').replace('linear0', 'u').replace(
            'linear1', 'v'
        ), rr[1]
        ) for rr in dimmap
    ]
    if verbose:
        print(f'dimmap: {dimmap}')
    if (
        ('linear0' in np.vstack(dimmap)[:,0])
        and ('linear1' in np.vstack(dimmap)[:,0])
    ):
        dimmap = [
            (rr[0].replace('linear0', 'l').replace('linear1', 'm'), rr[1])
            for rr in dimmap
        ]
    dimmap = [
        (rr[0].replace('linear0', 'component'), rr[1])
        for rr in dimmap if rr[1] >= 0
    ]
    # conversion to dict, example dimmap after this statement
    # dimmap: {'l': 0, 'm': 1, 'chan': 2, 'pol': 3}
    dimmap = dict(
        [(diraxes[int(rr[0][-1])], rr[1])
        if rr[0].startswith('linear') or rr[0].startswith('direction')
        else rr for rr in dimmap]
    )
    if verbose:
        print(f'dimmap: {dimmap}')
    return dimmap


def __get_freq_values(coords: coordinates.coordinatesystem, shape:tuple) -> list:
    idx = __get_image_axis_order(coords)[::-1].index('Frequency')
    if idx >= 0:
        coord_dict = coords.dict()
        for k in coord_dict:
            if k.startswith('spectral'):
                freqs = []
                wcs = coord_dict[k]['wcs']
                crpix = wcs['crpix']
                crval = wcs['crval']
                cdelt = wcs['cdelt']
                for i in range(shape[idx]):
                    f = (i - crpix) * cdelt + crval
                    freqs.append(f)
                return freqs
    else:
        return [1420e6]


def __get_image_axis_order(coords: coordinates.coordinatesystem) -> list:
    """
    get the *reverse* order of image axes
    """
    axis_names = coords.get_axes()[::-1]
    ncoords = len(axis_names)
    csys = coords.dict()
    ordered = len(__flatten_list(axis_names)) * ['']
    for i in range(ncoords-1, -1, -1):
        axes = csys['pixelmap' + str(i)]
        if len(axes) == 1:
            ordered[axes[0]] = axis_names[i]
        elif len(axes) == 2:
            ordered[axes[0]] = axis_names[i][1]
            ordered[axes[1]] = axis_names[i][0]
    ordered = __flatten_list(ordered)[::-1]
    return ordered


def __get_image_dim_order(coords: coordinates.coordinatesystem) -> list:
    """
    Get the xds dim order of the input image. The returned list is in casacore
    order, which is the reverse of the actual image dimension order
    """
    flat = __get_image_axis_order(coords)
    ret = []
    for axis in flat:
        b = axis.lower()
        if b.startswith('right') or b.startswith('uu'):
            ret.append('l')
        elif b.startswith('dec') or b.startswith('vv'):
            ret.append('m')
        elif b.startswith('freq'):
            ret.append('freq')
        elif b.startswith('stok'):
            ret.append('pol')
        else:
            raise Exception(f'Unhandled axis name {c}')
    return ret


def __get_mask_names(infile:str) -> list:
    t = tables.table(infile)
    tb_tool = tables.table(
        infile, readonly=True, lockoptions={'option': 'usernoread'}, ack=False
    )
    mymasks = t.getkeyword('masks') if 'masks' in t.keywordnames() else []
    t.close()
    return mymasks


def __get_multibeam(imageinfo: dict) -> Union[np.ndarray, None]:
    """Returns None if the image does not have multiple (per-plane) beams"""
    p = 'perplanebeams'
    if p not in imageinfo:
        return None
    beam = imageinfo[p]
    nchan = beam['nChannels']
    npol = beam['nStokes']
    beam_array = np.zeros([1, npol, nchan, 3])
    for c in range(nchan):
        for p in range(npol):
            k = nchan * p + c
            b = beam['*' + str(k)]
            beam_dict = __convert_beam_to_rad(b)
            beam_array[0][p][c][0] = beam_dict['major']['value']
            beam_array[0][p][c][1] = beam_dict['minor']['value']
            beam_array[0][p][c][2] = beam_dict['pa']['value']
    return beam_array


def __get_persistent_block(
    infile:str, shapes:tuple, starts:tuple, dimorder:list,
    transpose_list:list, new_axes:list
) -> xr.DataArray:
    block = __read_image_chunk(infile, shapes, starts)
    block = np.expand_dims(block, new_axes)
    block = block.transpose(transpose_list)
    block = xr.DataArray(block, dims=dimorder)
    return block


def __get_pol_values(coord_dict):
    for k in coord_dict:
        if k.startswith('stokes'):
            return coord_dict[k]['stokes']
    return ['I']


def __get_starts_shapes_slices(
    blockdes: dict, coords:coordinates.coordinatesystem, cshape:list
) -> tuple:
    img_dim_order = __get_image_dim_order(coords)
    starts = []
    shapes = []
    slices = {}
    for i, dim in enumerate(img_dim_order):
        if dim not in ['pol', 'freq', 'l', 'm', 'u', 'v']:
            raise Exception(f'Unsupported dimension {dim}')
        if dim in blockdes:
            extent = blockdes[dim]
            if isinstance(extent, int):
                starts.append(extent)
                shapes.append(extent+1)
                slices[dim] = slice(extent, extent+1)
            elif isinstance(extent, slice):
                starts.append(extent.start)
                shapes.append(extent.stop - extent.start)
                slices[dim] = slice(extent.start, extent.stop)
            else:
                raise Exception(f'Unhandled extent type {type(extent)}')
        else:
            starts.append(0)
            shapes.append(cshape[i])
    return starts, shapes, slices


def __get_time_values(coord_dict):
   return [ coord_dict['obsdate']['m0']['value'] ]


def __get_transpose_list(coords: coordinates.coordinatesystem) -> list:
    flat = __get_image_axis_order(coords)
    transpose_list = 5 * [-1]
    # time axis
    transpose_list[0] = 4
    new_axes = [4]
    last_axis = 3
    not_covered = ['l', 'm', 'u', 'v', 's', 'f']
    csys = coords.dict()
    for i, c in enumerate(flat):
        b = c.lower()
        if b.startswith('right') or b.startswith('uu'):
            transpose_list[3] = i
            # transpose_list[3] = csys['pixelmap0'][0]
            not_covered.remove('l')
            not_covered.remove('u')
        elif b.startswith('dec') or b.startswith('vv'):
            transpose_list[4] = i
            # transpose_list[4] = csys['pixelmap0'][1]
            not_covered.remove('m')
            not_covered.remove('v')
        elif b.startswith('freq'):
            # transpose_list[2] = csys['pixelmap1'][0]
            transpose_list[2] = i
            not_covered.remove('f')
        elif b.startswith('stok'):
            transpose_list[1] = i
            # transpose_list[1] = csys['pixelmap2'][0]
            not_covered.remove('s')
        else:
            raise Exception(f'Unhandled axis name {c}')
    h = {'l': 3, 'm': 4, 'u': 3, 'v': 4, 'f': 2, 's': 1}
    for p in not_covered:
        transpose_list[h[p]] = last_axis
        new_axes.append(last_axis)
        last_axis -= 1
    new_axes.sort()
    if transpose_list.count(-1) > 0:
        raise Exception(
            f'Logic error: axes {axes}, transpose_list {transpose_list}'
        )
    return transpose_list, new_axes


def __get_uv_values(coord_dict, axis_names, shape):
    for i, axis in enumerate(['UU', 'VV']):
        idx = axis_names.index(axis)
        if idx >= 0:
            for k in coord_dict:
                cdict = coord_dict[k]
                if k.startswith('linear'):
                    z = []
                    crpix = cdict['crpix'][i]
                    crval = cdict['crval'][i]
                    cdelt = cdict['cdelt'][i]
                    for i in range(shape[idx]):
                        f = (i - crpix) * cdelt + crval
                        z.append(f)
                    if axis == 'UU':
                        u = z
                    else:
                        v = z
    return u, v


def __get_velocity_values(coord_dict: dict, freq_values: list) -> list:
    restfreq = 1420405751.786
    for k in coord_dict:
        if k.startswith('spectral'):
            restfreq = coord_dict[k]['restfreq']
            break
    c = 2.99792458e+08
    # doppler type = RADIO definition
    return [ (1 - f/restfreq)*c for f in freq_values ]



def __make_coord_subset(xds:xr.Dataset, slices:dict) -> xr.Dataset:
    dim_to_coord_map = {}
    coord_to_dim_map = {}
    for c in xds.coords:
        coord_to_dim_map[c] = xds[c].dims
        for d in xds[c].dims:
            if c != d:
                if d in dim_to_coord_map:
                    dim_to_coord_map[d].append(c)
                else:
                    dim_to_coord_map[d] = [c]
    save_coord = {}
    for dim in slices:
        if dim in dim_to_coord_map:
            for coord_name in dim_to_coord_map[dim]:
                for dim_num, coord_dim in enumerate(xds[coord_name].dims):
                    if coord_dim == dim:
                        if coord_name not in save_coord:
                            save_coord[coord_name] = xds[coord_name].values
                        save_coord[coord_name] = np.take(
                            save_coord[coord_name],
                            range(slices[dim].start, slices[dim].stop),
                            dim_num
                        )
    xds = xds.drop_vars(save_coord.keys())
    for coord in xds.coords:
        if coord in slices:
            xds[coord] = xds[coord][slices[coord]]
    for coord in save_coord:
        xds = xds.assign_coords(
            {coord: (coord_to_dim_map[coord], save_coord[coord])}
        )
    return xds


def __multibeam_array(
        xds:xr.Dataset, img_full_path:str, as_dask_array:bool
    ) -> Union[xr.DataArray, None]:
    """This should only be called after the xds.beam attr has been set"""
    if xds.attrs['beam'] is None:
        # the image may have multiple beams
        casa_image = images.image(img_full_path)
        imageinfo = casa_image.info()['imageinfo']
        del casa_image
        mb = __get_multibeam(imageinfo)
        if mb is not None:
            # multiple beams are stored as a data varialbe, so remove
            # the beam xds attr
            del xds.attrs['beam']
            if as_dask_array:
                mb = da.array(mb)
            xdb = xr.DataArray(mb, dims=['time', 'pol', 'freq', 'beam_param'])
            xdb = xdb.rename('beam')
            xdb = xdb.assign_coords(beam_param=['major', 'minor', 'pa'])
            xdb.attrs['unit'] = 'rad'
            return xdb
    else:
        return None


###########################################################################
##
## read_generic_table() - read casacore table into memory resident xds
##
###########################################################################
def read_generic_table(infile, subtables=False, timecols=None, ignore=None):
    """
    read generic casacore table format to xarray dataset loaded in memory

    Parameters
    ----------
    infile : str
        Input table filename. To read a subtable simply append the subtable folder name under the main table (ie infile = '/path/mytable.tbl/mysubtable')
    subtables : bool
        Whether or not to include subtables underneath the specified table. If true, an attribute called subtables will be added to the returned xds.
        Default False
    timecols : list
        list of column names to convert to numpy datetime format. Default None leaves times as their original casacore format.
    ignore : list
        list of column names to ignore and not try to read. Default None reads all columns

    Returns
    -------
    xarray.core.dataset.Dataset
    """
    if timecols is None: timecols = []
    if ignore is None: ignore = []

    infile = os.path.expanduser(infile)
    assert os.path.isdir(infile), "invalid input filename to read_generic_table"

    attrs = extract_table_attributes(infile)
    tb_tool = tables.table(infile, readonly=True, lockoptions={'option': 'usernoread'}, ack=False)
    if tb_tool.nrows() == 0:
        tb_tool.close()
        return xr.Dataset(attrs=attrs)

    dims = ['row'] + ['d%i' % ii for ii in range(1, 20)]
    cols = tb_tool.colnames()
    ctype = dict([(col, tb_tool.getcell(col, 0)) for col in cols if (col not in ignore) and (tb_tool.iscelldefined(col, 0))])
    mvars, mcoords, xds = {}, {}, xr.Dataset()

    tr = tb_tool.row(ignore, exclude=True)[:]

    # extract data for each col
    for col in ctype.keys():
        if tb_tool.coldatatype(col) == 'record': continue  # not supported

        try:
            data = np.stack([rr[col] for rr in tr])  # .astype(ctype[col].dtype)
            if isinstance(tr[0][col], dict):
                data = np.stack([rr[col]['array'].reshape(rr[col]['shape']) if len(rr[col]['array']) > 0 else np.array(['']) for rr in tr])
        except:
            # sometimes the columns are variable, so we need to standardize to the largest sizes
            if len(np.unique([isinstance(rr[col], dict) for rr in tr])) > 1: continue  # can't deal with this case
            mshape = np.array(max([np.array(rr[col]).shape for rr in tr]))
            try:
                data = np.stack([np.pad(rr[col] if len(rr[col]) > 0 else np.array(rr[col]).reshape(np.arange(len(mshape)) * 0),
                                        [(0, ss) for ss in mshape - np.array(rr[col]).shape], 'constant', constant_values=np.array([np.nan]).astype(np.array(ctype[col]).dtype)[0]) for rr in tr])
            except:
                data = []

        if len(data) == 0: continue
        if col in timecols: convert_time(data)
        if col.endswith('_ID'):
            mcoords[col] = xr.DataArray(data, dims=['d%i_%i' % (di, ds) for di, ds in enumerate(np.array(data).shape)])
        else:
            mvars[col] = xr.DataArray(data, dims=['d%i_%i' % (di, ds) for di, ds in enumerate(np.array(data).shape)])

    xds = xr.Dataset(mvars, coords=mcoords)
    xds = xds.rename(dict([(dv, dims[di]) for di, dv in enumerate(xds.dims)]))
    attrs['bad_cols'] = list(np.setdiff1d([dv for dv in tb_tool.colnames()], [dv for dv in list(xds.data_vars) + list(xds.coords)]))

    # if this table has subtables, use a recursive call to store them in subtables attribute
    if subtables:
        stbl_list = sorted([tt for tt in os.listdir(infile) if os.path.isdir(os.path.join(infile, tt)) and tables.tableexists(os.path.join(infile, tt))])
        attrs['subtables'] = []
        for ii, subtable in enumerate(stbl_list):
            sxds = read_generic_table(os.path.join(infile, subtable), subtables=subtables, timecols=timecols, ignore=ignore)
            if len(sxds.dims) != 0: attrs['subtables'] += [(subtable, sxds)]

    xds = xds.assign_attrs(attrs)
    tb_tool.close()

    return xds


def __read_image_array(
    infile:str, chunks:Union[list, dict], mask:str=None, verbose:bool=False,
    blc=None, trc=None
) -> dask.array:
    """
    Read an array of image pixels into a dask array. The returned dask array
    will have axes in time, pol, freq, l, m order
    If specified, it's the caller's responsibility to ensure the specified blc and
    trc are coincident with chunk corners. If not, there could be performance degradation
    The blc, trc box is inclusive of blc pixel coordinates and exclusive trc pixel
    coordinates, which mimics the behavior of numpy array slicing
    ie blc = [0, 0, 0] and trc = [1023, 1023, 1023] excludes pixel [1023, 1023, 1023]
    and has shape [1023, 1023, 1023]
    blc and trc can be any array type object
    blc and trc are taken wrt the CASA image. Not the casacore image in which they
    would be reversed
    :param infile: Path to the input CASA image
    :type infile: str, required
    :param chunks: The desired dask chunk size.
                   If list, the ordering is based on the ordering of the input
                   axes. So, for example, if the input image has axes RA, Dec,
                   Freq, Stokes and the desired chunking is 40 pixels in RA, 30
                   pixels in Dec, 20 pixels in Freq, and 2 pixels in Stokes,
                   chunks would be specified as [40, 30, 20, 2].
                   If dict, supported optional keys are 'l', 'm', 'freq', 'pol',
                   and 'time'. The supported values are positive integers,
                   indicating the length of a chunk on that particular axis. If
                   a key is missing, the associated chunk length along that axis
                   is 1. 'l' represents the longitude like dimension, and 'm'
                   represents the latitude like dimension. For apeature images,
                   'u' may be used in place of 'l', and 'v' in place of 'm'.
    :type chunks: list | dict, required
    :param mask: If specified, this is the associated image mask to read, rather than the actual
                 image pixel values. None means read the image pixel values, not the mask.
    :type mask: str, optional
    :param verbose: Emit debugging messages, default False
    :type verbose: bool
    :param blc: bottom left corner, given in the axes ordering of the input image.None=>The origin
    :type blc: a type that is convertable to a list via list(blc)
    :param trc: top right corner, given in the axes ordering of the input image.None=>image shape - 1
    :type trc: a type that is convertable to a list via list(trc)
    :return: Dask array in time, pol, freq, l, m order
    :rtype: dask.array

    """
    img_full_path = os.path.expanduser(infile)
    casa_image = images.image(img_full_path)
    if isinstance(chunks, list):
        mychunks = tuple(chunks)
    elif isinstance(chunks, dict):
        mychunks = __get_chunk_list(
            chunks, casa_image.coordinates().get_names()[::-1], casa_image.shape()[::-1]
        )
    else:
        raise Exception(
            f'incorrect type {type(chunks)} for parameter chunks. Must be '
            'either tuple or dict'
        )
    # shape list is the reverse of the actual image shape
    cshape = casa_image.shape()
    transpose_list, new_axes = __get_transpose_list(casa_image.coordinates())
    data_type = casa_image.datatype()
    del casa_image
    if verbose:
        print(f'cshape: {cshape}')
    if mask:
        img_full_path = os.sep.join([img_full_path, mask])
    # rblc and rtrc are enforced to be of type list
    if blc is None:
        rblc = len(cshape) * [0]
    else:
        if len(blc) == len(cshape):
            rblc = list(blc[::-1])
            if rblc >= cshape:
                raise RuntimeError(f'blc {blc} >= shape {cshape[::-1]}')
        else:
            raise RuntimeError(
                f'cshape has {len(cshape)} dimensions and blc has {len(blc)}'
            )
    rblc = rblc + [0 for rr in range(5) if rr >= len(mychunks)]
    if trc is None:
        rtrc = cshape
    else:
        if len(trc) == len(cshape):
            rtrc = list(trc[::-1])
            if rtrc > cshape:
                raise RuntimeError(f'trc {trc} >= shape {cshape[::-1]}')
            xblc = rblc if blc is None else list(blc)
            if list(trc) <= xblc:
                raise RuntimeError(f'trc {trc} must be > blc {blc}')
        else:
            raise RuntimeError(
                f'cshape has {len(cshape)} dimensions and blc has {len(blc)}'
            )
    rtrc = rtrc + [1 for rr in range(5) if rr >= len(mychunks)]

    # expand the actual data shape to the full 5 possible dims
    # full_shape = cshape + [1 for rr in range(5) if rr >= len(cshape)]
    # full_shape = list(rtrc - rblc)
    full_chunks = mychunks[::-1] + tuple([1 for rr in range(5) if rr >= len(mychunks)])
    d0slices = []
    for d0 in range(rblc[0], rtrc[0], full_chunks[0]):
        d0len = min(full_chunks[0], rtrc[0] - d0)
        d1slices = []

        for d1 in range(rblc[1], rtrc[1], full_chunks[1]):
            d1len = min(full_chunks[1], rtrc[1] - d1)
            d2slices = []

            for d2 in range(rblc[2], rtrc[2], full_chunks[2]):
                d2len = min(full_chunks[2], rtrc[2] - d2)
                d3slices = []

                for d3 in range(rblc[3], rtrc[3], full_chunks[3]):
                    d3len = min(full_chunks[3], rtrc[3] - d3)
                    d4slices = []

                    for d4 in range(rblc[4], rtrc[4], full_chunks[4]):
                        d4len = min(full_chunks[4], rtrc[4] - d4)

                        shapes = tuple(
                            [d0len, d1len, d2len, d3len, d4len][:len(cshape)]
                        )
                        starts = tuple([d0, d1, d2, d3, d4][:len(cshape)])
                        delayed_array = dask.delayed(__read_image_chunk)(
                            img_full_path, shapes, starts
                        )
                        d4slices += [
                            dask.array.from_delayed(
                                delayed_array, shapes, data_type
                            )
                        ]
                    d3slices += (
                        [dask.array.concatenate(d4slices, axis=4)]
                        if len(cshape) > 4 else d4slices
                    )
                d2slices += (
                    [dask.array.concatenate(d3slices, axis=3)]
                    if len(cshape) > 3 else d3slices
                )
            d1slices += (
                [dask.array.concatenate(d2slices, axis=2)]
                if len(cshape) > 2 else d2slices
            )
        d0slices += (
            [dask.array.concatenate(d1slices, axis=1)]
            if len(cshape) > 1 else d1slices
        )
    ary = dask.array.concatenate(d0slices, axis=0)
    ary = da.expand_dims(ary, new_axes)
    return ary.transpose(transpose_list)


def __read_image_chunk(infile:str, shapes:tuple, starts:tuple) -> np.ndarray:
    tb_tool: tables.table = tables.table(
        infile, readonly=True, lockoptions={'option': 'usernoread'}, ack=False
    )
    data : np.ndarray = tb_tool.getcellslice(
        tb_tool.colnames()[0], 0, starts,
        tuple(np.array(starts) + np.array(shapes) - 1)
    )
    tb_tool.close()
    return data





import os
import shutil
import requests
import zipfile

from tqdm import tqdm
#
#https://www.dropbox.com/scl/fi/x8tp0wu21gssbd1gxrnmy/Antennae_North.cal.lsrk.vis.zarr.zip?rlkey=l9jdr6tvyq4pe3381gukuly0d&dl=0

FILE_ID = {
  'demo_simulated.im':
  {
    'file':'demo_simulated.im.zip',
    'id':'z87gibxshwg9e2h155ukk',
    'rlkey':'bn7uvs697wtedij63fa2hu7ed&dl'
  },
 'Antennae_North.cal.lsrk.vis.zarr':
  {
    'file':'Antennae_North.cal.lsrk.vis.zarr.zip',
    'id':'x8tp0wu21gssbd1gxrnmy',
    'rlkey':'l9jdr6tvyq4pe3381gukuly0d&dl'
  },
  
        
}

def download(file, folder='.'):
  if os.path.exists('/'.join((folder, file))):
    print("File exists.")
    return
    
  if file not in FILE_ID.keys():
    print("Requested file not found")
    
    return 
  
  fullname=FILE_ID[file]['file']
  id=FILE_ID[file]['id']
  rlkey=FILE_ID[file]['rlkey']
    
  url = 'https://www.dropbox.com/scl/fi/{id}/{file}?rlkey={rlkey}'.format(id=id, file=fullname, rlkey=rlkey)
    
  headers = {'user-agent': 'Wget/1.16 (linux-gnu)'}

  r = requests.get(url, stream=True, headers=headers)
  total = int(r.headers.get('content-length', 0))

  fullname = '/'.join((folder, fullname))

  with open(fullname, 'wb') as fd, tqdm(
    desc=fullname,
    total=total,
    unit='iB',
    unit_scale=True,
    unit_divisor=1024) as bar:
      for chunk in r.iter_content(chunk_size=1024):
        if chunk:
          size=fd.write(chunk)
          bar.update(size)
                
  if zipfile.is_zipfile(fullname):                
    shutil.unpack_archive(filename=fullname, extract_dir=folder)
    
    # Let's clean up after ourselves
    os.remove(fullname)



#  CASA Next Generation Infrastructure
#  Copyright (C) 2021, 2023 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# From the casacore::Stokes class definition
stokes_types = {
    0: "Undefined",
    1: "I",
    2: "Q",
    3: "U",
    4: "V",
    5: "RR",
    6: "RL",
    7: "LR",
    8: "LL",
    9: "XX",
    10: "XY",
    11: "YX",
    12: "YY",
    13: "RX",
    14: "RY",
    15: "LX",
    16: "LY",
    17: "XR",
    18: "XL",
    19: "YR",
    20: "YL",
    21: "PP",
    22: "PQ",
    23: "QP",
    24: "QQ",
    25: "RCircular",
    26: "LCircular",
    27: "Linear",
    28: "Ptotal",
    29: "Plinear",
    30: "PFtotal",
    31: "PFlinear",
    32: "Pangle",
}



#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#################################
# Helper File
#
# Not exposed in API
#
#################################
import warnings, time, os, logging
import numpy as np
import astropy.wcs
from .common import __get_xds_dim_order, __dask_arrayize
from ._casacore.common import __active_mask
from ._casacore.xds_to_casacore import (
    __coord_dict_from_xds, __history_from_xds, __imageinfo_dict_from_xds,
    __write_casa_data
)
from ._casacore.xds_from_casacore import (
    __add_coord_attrs, __add_mask, __add_sky_or_apeture,
    __casa_image_to_xds_attrs, __casa_image_to_xds_metadata, __get_mask_names,
    __get_persistent_block, __get_starts_shapes_slices, __get_transpose_list,
    __make_coord_subset, __multibeam_array, __read_image_array
)
from astropy.coordinates import SkyCoord  # High-level coordinates
from astropy.coordinates import ICRS, Galactic, FK4, FK5  # Low-level frames
from astropy.coordinates import Angle, Latitude, Longitude  # Angles
from astropy import units as u
from casacore import images, tables
from typing import Union
import xarray as xr
from casacore.images import image

warnings.filterwarnings('ignore', category=FutureWarning)


def __load_casa_image_block(infile: str, block_des: dict) -> xr.Dataset:
    image_full_path = os.path.expanduser(infile)
    casa_image = image(image_full_path)
    coords = casa_image.coordinates()
    cshape = casa_image.shape()
    del casa_image
    ret = __casa_image_to_xds_metadata(image_full_path, False)
    xds = ret['xds']
    starts, shapes, slices = __get_starts_shapes_slices(block_des, coords, cshape)
    xds = __make_coord_subset(xds, slices)
    dimorder = __get_xds_dim_order(ret['sphr_dims'])
    transpose_list, new_axes = __get_transpose_list(coords)
    block = __get_persistent_block(image_full_path, shapes, starts, dimorder, transpose_list, new_axes)
    xds = __add_sky_or_apeture(xds, block, dimorder, image_full_path, ret['sphr_dims'])
    mymasks = __get_mask_names(image_full_path)
    for m in mymasks:
        full_path = os.sep.join([image_full_path, m])
        block = __get_persistent_block(
            full_path, shapes, starts, dimorder, transpose_list, new_axes
        )
        xds = __add_mask(xds, m, block, dimorder)
    xds.attrs = __casa_image_to_xds_attrs(image_full_path, True)
    mb = __multibeam_array(xds, image_full_path, False)
    if mb is not None:
        selectors = {}
        for k in ('time', 'pol', 'freq'):
            if k in block_des:
                selectors[k] = block_des[k]
        xds['beam'] = mb.isel(selectors)
    xds = __add_coord_attrs(xds, ret['icoords'], ret['dir_axes'])
    return xds


def __read_casa_image(
    infile:str, chunks:Union[list, dict], masks:bool=True,
    history:bool=True, verbose:bool=False
) -> xr.Dataset:
    img_full_path = os.path.expanduser(infile)
    ret = __casa_image_to_xds_metadata(img_full_path, verbose)
    xds = ret['xds']
    dimorder = __get_xds_dim_order(ret['sphr_dims'])
    xds = __add_sky_or_apeture(
        xds, __read_image_array(img_full_path, chunks, verbose=verbose),
        dimorder, img_full_path, ret['sphr_dims']
    )
    if masks:
        mymasks = __get_mask_names(img_full_path)
        for m in mymasks:
            ary = __read_image_array(img_full_path, chunks, mask=m, verbose=verbose)
            xds = __add_mask(xds, m, ary, dimorder)
    xds.attrs = __casa_image_to_xds_attrs(img_full_path, history)
    mb = __multibeam_array(xds, img_full_path, True)
    if mb is not None:
        xds['beam'] = mb
    xds = __add_coord_attrs(xds, ret['icoords'], ret['dir_axes'])
    xds = __dask_arrayize(xds)
    return xds


def __xds_to_casa_image(xds: xr.Dataset, imagename:str) -> None:
    sky_ap = 'sky' if 'sky' in xds else 'apeature'
    if xds[sky_ap].shape[0] != 1:
        raise Exception('XDS can only be converted if it has exactly one time plane')
    arr = xds[sky_ap].isel(time=0).transpose(*('freq', 'pol', 'm', 'l'))
    image_full_path = os.path.expanduser(imagename)
    maskname = xds.active_mask if __active_mask in xds.attrs else ''
    # create the image and then delete the object
    casa_image = images.image(image_full_path, maskname=maskname, shape=arr.shape)
    del casa_image
    chunk_bounds = arr.chunks
    __write_casa_data(xds, image_full_path, arr.shape[::-1])
    # create coordinates
    coord = __coord_dict_from_xds(xds)
    ii = __imageinfo_dict_from_xds(xds)
    units = xds.sky.attrs['unit'] if 'unit' in xds.sky.attrs else None
    miscinfo = (
        xds.attrs['user']
        if 'user' in xds.attrs and len(xds.attrs['user']) > 0 else None
    )
    tb = tables.table(
        image_full_path, readonly=False, lockoptions={'option': 'permanentwait'},
        ack=False
    )
    tb.putkeyword('coord', coord)
    tb.putkeyword('imageinfo', ii)
    if units:
        tb.putkeyword('units', units)
    if miscinfo:
        tb.putkeyword('miscinfo', miscinfo)
    tb.done()
    # history
    __history_from_xds(xds, image_full_path)



from astropy.wcs import WCS
import numpy as np
import xarray as xr
from typing import Union
from .common import __c

def __make_empty_sky_image(
    xds:xr.Dataset, phase_center:Union[list, np.ndarray],
	image_size:Union[list, np.ndarray], cell_size:Union[list, np.ndarray],
	chan_coords:Union[list, np.ndarray],
	pol_coords:Union[list, np.ndarray], time_coords:Union[list, np.ndarray],
    direction_reference:str, projection:str, spectral_reference:str
) -> xr.Dataset:
    if len(image_size) != 2:
        raise ValueError('image_size must have exactly two elements')
    if len(phase_center) != 2:
        raise ValueError('phase_center must have exactly two elements')
    if len(cell_size) != 2:
        raise ValueError('cell_size must have exactly two elements')
    wcs_dict = {}
    wcs_dict['NAXIS1'] = image_size[0]
    wcs_dict['CTYPE1'] = f'RA---{projection}'
    wcs_dict['CRVAL1'] = phase_center[0]
    wcs_dict['CRPIX1'] = image_size[0]//2 + 1
    wcs_dict['CDELT1'] = -abs(cell_size[0])
    wcs_dict['CUNIT1'] = 'rad'
    wcs_dict['NAXIS2'] = image_size[1]
    wcs_dict['CTYPE2'] = f'DEC--{projection}'
    wcs_dict['CRVAL2'] = phase_center[1]
    wcs_dict['CRPIX2'] = image_size[1]//2 + 1
    wcs_dict['CDELT2'] = abs(cell_size[1])
    wcs_dict['CUNIT2'] = 'rad'
    w = WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    long, lat = w.pixel_to_world_values(x, y)
    # long, lat from above eqn will always be in degrees, so convert to rad
    f = np.pi/180
    long *= f
    lat *= f
    if not isinstance(chan_coords, list) and not isinstance(chan_coords, np.ndarray):
        chan_coords = [ chane_coords ]
    chan_coords = np.array(chan_coords, dtype=np.float64)
    restfreq = chan_coords[len(chan_coords)//2]
    vel = (1 - chan_coords/restfreq) * __c
    if not isinstance(time_coords, list) and not isinstance(time_coords, np.ndarray):
        time_coords = [ time_coords ]
    time_coords = np.array(time_coords, dtype=np.float64)
    coords = {
        'time': time_coords, 'pol': pol_coords, 'freq': chan_coords,
        'vel': (('freq'), vel), 'right_ascension': (('l', 'm'), long),
        'declination': (('l', 'm'), lat)
    }
    xds = xds.assign_coords(coords)
    xds.time.attrs = {'format': 'MJD', 'refer': 'UTC', 'unit': 'd'}
    xds.freq.attrs = {
        'conversion': {
            'direction': {
                'm0': {'unit': 'rad', 'value': 0.0},
                'm1': {'unit': 'rad', 'value': 1.5707963267948966},
                'refer': 'FK5',
                'type': 'direction'
            },
            'epoch': {
                'm0': {'unit': 'd', 'value': 0.0}, 'refer': 'LAST',
                'type': 'epoch'
            },
            'position': {
                'm0': {'unit': 'rad', 'value': 0.0},
                'm1': {'unit': 'rad', 'value': 0.0},
                'm2': {'unit': 'm', 'value': 0.0},
                'refer': 'ITRF',
                'type': 'position'
            },
            'system': spectral_reference.upper()
        },
        'native_type': 'FREQ',
        'restfreq': restfreq,
        'restfreqs': [ restfreq ],
        'system': spectral_reference.upper(),
        'unit': 'Hz',
        'wave_unit': 'mm',
        'wcs': {
            'crval': chan_coords[len(chan_coords)//2],
            'cdelt': chan_coords[1] - chan_coords[0] if len(chan_coords) > 1 else 1000.0,
            'pc': 1.0
        }
    }
    xds.vel.attrs = {'doppler_type': 'RADIO', 'unit': 'm/s'}
    xds.right_ascension.attrs = {
        'unit': 'rad',
        'wcs': {
            'crval': phase_center[0], 'cdelt': -abs(cell_size[0])
        }
    }
    xds.declination.attrs = {
        'unit': 'rad',
        'wcs': {
            'crval': phase_center[1], 'cdelt': abs(cell_size[1])
        }
    }
    xds.attrs = {
        'direction': {
            'conversion_system': direction_reference,
            'conversion_equinox': 'J2000',
            'long_pole': 0.0, 'lat_pole': 0.0,
            'pc': np.array([[1.0, 0.0], [0.0, 1.0]]),
            'projection': projection,
            'projection_parameters': np.array([0.0, 0.0]),
            'system': direction_reference, 'equinox': 'J2000',
        },
        'active_mask': '', 'beam': None, 'object_name': '',
        'obsdate': {
            'refer': 'UTC', 'format': 'MJD',
            'value': time_coords[0], 'unit': 'd'
        },
        'observer': 'Karl Jansky',
        'pointing_center': {
            'value': np.array(phase_center), 'initial': True
        },
        'description': '',
        'telescope': {
            'name': 'ALMA',
            'position': {
                'type': 'position', 'refer': 'ITRF',
                'm2': {'value': 6379946.01326443, 'unit': 'm'},
                'm1': {'unit': 'rad', 'value': -0.3994149869262738},
                'm0': {'unit': 'rad', 'value': -1.1825465955049892}
            }
        },
        'history': None,
    }
    return xds



#  CASA Next Generation Infrastructure
#  Copyright (C) 2023 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Generator
from contextlib import contextmanager

from casacore import tables


@contextmanager
def open_table_ro(infile: str) -> Generator[tables.table, None, None]:
    table = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    try:
        yield table
    finally:
        table.close()


@contextmanager
def open_query(table: tables.table, query: str) -> Generator[tables.table, None, None]:
    ttq = tables.taql(query)
    try:
        yield ttq
    finally:
        ttq.close()



import numpy as np


__np_types = {
    'complex128': np.complex128, 'complex256': np.complex256,
    'complex64': np.complex64,
    'float128': np.float128, 'float64': np.float64,
    'float16': np.float16, 'float32': np.float32,
    'int8': np.int8, 'int16': np.int16, 'int32': np.int32, 'int64': np.int64
}


__top_level_sub_xds = '__attrs_xds'




__active_mask = 'active_mask'
__doppler_types = ['RADIO', 'Z', 'RATIO', 'BETA', 'GAMMA']
__native_types = ['FREQ', 'VRAD', 'VOPT', 'BETA', 'WAVE', 'AWAV']
__object_name = 'object_name'
__pointing_center = 'pointing_center'




#  CASA Next Generation Infrastructure
#  Copyright (C) 2021, 2023 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Dict, Union

import numpy as np
import pandas as pd

from ._tables.read import read_generic_table, read_flat_col_chunk
from ._tables.table_query import open_query, open_table_ro


def describe_ms(
    infile: str, mode="summary", rowmap: Union[dict, None] = None
) -> Union[pd.DataFrame, Dict]:
    """
    Summarize the contents of an MS directory in casacore table format

    :param infile: input MS filename
    :param mode: type of information returned ('summary', 'flat', 'expanded').
    'summary' returns a pandas dataframe that is nice for displaying in notebooks
    etc. 'flat' returns a list of tuples of (ddi, row, chan, pol). 'expanded'
    returns a list of tuples of (ddi, time, baseline, chan, pol). These latter two
    are good for trying to determine chunk size for read_ms(expand=True/False).
    :param rowmap: dict of DDI to tuple of (row indices, channel indices). Returned
    by ms_selection function. Default None ignores selections

    :return: summary as a pd dataframe
    """
    infile = os.path.expanduser(infile)  # does nothing if $HOME is unknown
    if not os.path.isdir(infile):
        raise ValueError(f"invalid input filename to describe_ms: {infile}")
    if mode not in [
        "summary",
        "flat",
        "expanded",
    ]:
        raise ValueError("invalid mode, must be summary, flat or expanded")

    ddi_xds = read_generic_table(infile, "DATA_DESCRIPTION")
    ddis = list(ddi_xds.row.values) if rowmap is None else list(rowmap.keys())

    summary: Union[pd.DataFrame, Dict] = []
    if mode == "summary":
        summary = pd.DataFrame([])

    with open_table_ro(infile) as tb_tool:
        for ddi in ddis:
            taql = f"select * from $tb_tool where DATA_DESC_ID = {ddi}"
            with open_query(tb_tool, taql) as query_per_ddi:
                populate_ms_descr(infile, mode, query_per_ddi, summary, ddi, ddi_xds)

    if mode == "summary":
        summary = summary.set_index("ddi").sort_index()
    else:
        summary = dict(summary)
    return summary


def populate_ms_descr(
    infile, mode, query_per_ddi, summary, ddi, ddi_xds, rowmap=None
) -> pd.DataFrame:
    """Adds information from the time and baseline (antenna1+antenna2)
    columns as well as channel and polarizations, based on a taql
    query.

    :param mode: mode (as in describe_ms())
    :param query_per_ddi: a TaQL query with data per individual DDI
    :param sdf: summary dict being populated
    :param summary: final summary object being populated from the invividual sdf's
    """
    spw_ids = ddi_xds.spectral_window_id.values
    pol_ids = ddi_xds.polarization_id.values
    sdf = {
        "ddi": ddi,
        "spw_id": spw_ids[ddi],
        "pol_id": pol_ids[ddi],
        "rows": query_per_ddi.nrows(),
    }

    # figure out characteristics of main table from select subtables (must all be present)
    spw_xds = read_generic_table(infile, "SPECTRAL_WINDOW")
    pol_xds = read_generic_table(infile, "POLARIZATION")

    if mode in ["expanded", "summary"]:
        times = (
            query_per_ddi.getcol("TIME")
            if rowmap is None
            else read_flat_col_chunk(infile, "TIME", (1,), rowmap[ddi][0], 0, 0)
        )
        baselines = [
            query_per_ddi.getcol(rr)[:, None]
            if rowmap is None
            else read_flat_col_chunk(infile, rr, (1,), rowmap[ddi][0], 0, 0)
            for rr in ["ANTENNA1", "ANTENNA2"]
        ]
        sdf.update(
            {
                "times": len(np.unique(times)),
                "baselines": len(np.unique(np.hstack(baselines), axis=0)),
            }
        )

    chans = spw_xds.num_chan.values
    pols = pol_xds.num_corr.values
    sdf.update(
        {
            "chans": chans[spw_ids[ddi]]
            if (rowmap is None) or (rowmap[ddi][1] is None)
            else len(rowmap[ddi][1]),
            "pols": pols[pol_ids[ddi]],
        }
    )
    sdf["size_MB"] = np.ceil(
        (sdf["rows"] * sdf["chans"] * sdf["pols"] * 10) / 1024**2
    ).astype(int)

    if rowmap is not None:
        sdf["rows"] = len(rowmap[ddi][0])

    if mode == "summary":
        summary = pd.concat(
            [summary, pd.DataFrame(sdf, index=[str(ddi)])], axis=0, sort=False
        )
    elif mode == "flat":
        summary += [(ddi, (sdf["rows"], sdf["chans"], sdf["pols"]))]
    else:
        summary += [((ddi, sdf["times"], sdf["baselines"], sdf["chans"], sdf["pols"]))]

    return sdf



import astropy
from astropy.coordinates import Angle, SkyCoord
from casacore import tables
import numpy as np
import os
import xarray as xr
from .common import (
    __active_mask, __doppler_types, __native_types, __object_name,
    __pointing_center
)


def __compute_direction_dict(xds: xr.Dataset) -> dict:
    """
    Given xds metadata, compute the direction dict that is valid
    for a CASA image coordinate system
    """
    direction = {}
    xds_dir = xds.attrs['direction']
    direction['system'] = xds_dir['equinox']
    direction['projection'] = xds_dir['projection']
    direction['projection_parameters'] = xds_dir['projection_parameters']
    long = xds.right_ascension
    lat = xds.declination
    direction['units'] = np.array(
        [long.attrs['unit'], lat.attrs['unit']], dtype='<U16'
    )
    direction['crval'] = np.array([
        long.attrs['wcs']['crval'], lat.attrs['wcs']['crval']
    ])
    direction['cdelt'] = np.array([
        long.attrs['wcs']['cdelt'], lat.attrs['wcs']['cdelt']
    ])
    # get the actual ref pix
    proj = direction['projection']
    wcs_dict = {}
    wcs_dict[f'CTYPE1'] = f'RA---{proj}'
    wcs_dict[f'NAXIS1'] = long.shape[0]
    wcs_dict[f'CUNIT1'] = long.attrs['unit']
    # FITS arrays are 1-based
    wcs_dict[f'CRPIX1'] = 1
    wcs_dict[f'CRVAL1'] = long[0][0].item(0)
    wcs_dict[f'CDELT1'] = long.attrs['wcs']['cdelt']
    wcs_dict[f'CTYPE2'] = f'DEC--{proj}'
    wcs_dict[f'NAXIS2'] = lat.shape[1]
    wcs_dict[f'CUNIT2'] = lat.attrs['unit']
    # FITS arrays are 1-based
    wcs_dict[f'CRPIX2'] = 1
    wcs_dict[f'CRVAL2'] = lat[0][0].item(0)
    wcs_dict[f'CDELT2'] = lat.attrs['wcs']['cdelt']
    w = astropy.wcs.WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    sky = SkyCoord(direction[
        'crval'][0], direction['crval'][1],
        frame=xds.attrs['direction']['system'].lower(),
        equinox=xds.attrs['direction']['equinox'],
        unit=long.attrs['unit']
    )
    crpix = w.world_to_pixel(sky)
    direction['crpix'] = np.array([crpix[0], crpix[1]])
    direction['pc'] = xds_dir['pc']
    direction['axes'] = ['Right Ascension', 'Declination']
    direction['conversionSystem'] = direction['system']
    for s in ['longpole', 'latpole']:
        direction[s] = Angle(
            str(xds_dir[s]['value']) + xds_dir[s]['unit']
        ).deg
    return direction


def __compute_spectral_dict(
    xds: xr.Dataset, direction: dict, obsdate: dict, tel_pos: dict,
) -> dict:
    """
    Given xds metadata, compute the spectral dict that is valid
    for a CASA image coordinate system
    """
    spec = {}
    spec_conv = {}
    spec_conv['direction'] = {
        'm0': {'unit': direction['units'][0], 'value': direction['crval'][0]},
        'm1': {'unit': direction['units'][1], 'value': direction['crval'][1]},
        'refer': direction['system'], 'type': 'direction'
    }
    spec_conv['epoch'] = obsdate
    spec_conv['position'] = tel_pos
    spec_conv['system'] = xds.freq.attrs['system']
    spec['conversion'] = spec_conv
    spec['formatUnit'] = ''
    spec['name']: 'Frequency'
    spec['nativeType'] = __native_types.index(xds.freq.attrs['native_type'])
    spec['restfreq'] = xds.freq.attrs['restfreq']
    spec['restfreq'] = xds.freq.attrs['restfreqs']
    spec['system'] = xds.freq.attrs['system']
    spec['unit'] = xds.freq.attrs['unit']
    spec['velType'] = __doppler_types.index(xds.vel.attrs['doppler_type'])
    spec['velUnit'] = xds.vel.attrs['unit']
    spec['version'] = 2
    spec['waveUnit'] = xds.freq.attrs['wave_unit']
    spec_wcs = xds.freq.attrs['wcs']
    # spec_wcs['ctype'] = 'FREQ'
    # spec_wcs['pc'] = 1.0
    # spec_wcs['crpix'] = (spec_wcs['crval'] - xds.freq.values[0])/spec_wcs['cdelt']
    spec['wcs'] = spec_wcs
    return spec


def __coord_dict_from_xds(xds: xr.Dataset) -> dict:
    coord = {}
    coord['telescope'] = xds.attrs['telescope']['name']
    coord['observer'] = xds.attrs['observer']
    obsdate = {}
    obsdate['refer'] = xds.time.attrs['refer']
    obsdate['unit'] = xds.time.attrs['unit']
    obsdate['value'] = xds.time.values[0]
    obsdate['format'] = xds.time.attrs['format']
    coord['obsdate'] = obsdate
    coord['pointingcenter'] = xds.attrs[__pointing_center].copy()
    coord['telescopeposition'] = xds.attrs['telescope']['position'].copy()
    coord['direction0'] = __compute_direction_dict(xds)
    coord['stokes1'] = {
        'axes': np.array(['Stokes'], dtype='<U16'), 'cdelt': np.array([1.]),
        'crpix': np.array([0.]), 'crval': np.array([1.]), 'pc': np.array([[1.]]),
        'stokes': np.array(xds.pol.values, dtype='<U16')
    }
    coord['spectral2'] = __compute_spectral_dict(
        xds, coord['direction0'], coord['obsdate'], coord['telescopeposition']
    )
    coord['pixelmap0'] = np.array([0, 1])
    coord['pixelmap1'] = np.array([2])
    coord['pixelmap2'] = np.array([3])
    coord['pixelreplace0'] = np.array([0., 0.])
    coord['pixelreplace1'] = np.array([0.])
    coord['pixelreplace2'] = np.array([0.])
    coord['worldmap0'] = np.array([0, 1])
    coord['worldmap1'] = np.array([2])
    coord['worldmap2'] = np.array([3])
    coord['worldreplace0'] = coord['direction0']['crval']
    coord['worldreplace1'] = np.array(coord['stokes1']['crval'])
    coord['worldreplace2'] = np.array([xds.freq.attrs['wcs']['crval']])
    return coord


def __history_from_xds(xds: xr.Dataset, image: str) -> None:
    tb = tables.table(
        os.sep.join([image, 'logtable']), readonly=False,
        lockoptions={'option': 'permanentwait'}, ack=False
    )
    nrows = len(xds.history.row) if 'row' in xds.data_vars else 0
    if nrows > 0:
        # TODO need to implement nrows == 0 case
        tb.addrows(nrows + 1)
        for c in ['TIME', 'PRIORITY', 'MESSAGE', 'LOCATION', 'OBJECT_ID']:
            vals = xds.history[c].values
            if c == 'TIME':
                k = time.time() + 40587*86400
            elif c == 'PRIORITY':
                k = 'INFO'
            elif c == 'MESSAGE':
                k = (
                    'Wrote xds to ' + os.path.basename(image)
                    + ' using cngi_io.xds_to_casa_image_2()'
                )
            elif c == 'LOCATION':
                k = 'cngi_io.xds_to_casa_image_2'
            elif c == 'OBJECT_ID':
                k = ''
            vals = np.append(vals, k)
            tb.putcol(c, vals)
    tb.close()


def __imageinfo_dict_from_xds(xds: xr.Dataset) -> dict:
    ii = {}
    ii['image_type'] = (
        xds.sky.attrs['image_type'] if 'image_type' in xds.sky.attrs else ''
    )
    ii['objectname'] = xds.attrs[__object_name]
    if 'beam' in xds.attrs:
        if xds.attrs['beam']:
            # do nothing if xds.attrs['beam'] is None
            ii['restoringbeam'] = xds.attrs['beam']
    elif 'beam' in xds.data_vars:
        # multi beam
        pp = {}
        pp['nChannels'] = len(xds.freq)
        pp['nStokes'] = len(xds.pol)
        bu = xds.beam.attrs['unit']
        chan = 0
        pol = 0
        bv = xds.beam.values
        for i in range(pp['nChannels'] * pp['nStokes']):
            bp = bv[0][pol][chan][:]
            b = {
                'major': {'unit': bu, 'value': bp[0]},
                'minor': {'unit': bu, 'value': bp[1]},
                'positionangle': {'unit': bu, 'value': bp[2]}
            }
            pp['*' + str(pp['nChannels']*pol + chan)] = b
            chan += 1
            if chan >= pp['nChannels']:
                chan = 0
                pol += 1
        ii['perplanebeams'] = pp
    return ii


def __write_casa_data(xds: xr.Dataset, image_full_path: str, casa_image_shape: tuple) -> None:
    sky_ap = 'sky' if 'sky' in xds else 'apeature'
    active_mask = xds.active_mask if __active_mask in xds.attrs else ''
    masks = []
    masks_rec = {}
    for m in xds.data_vars:
        attrs = xds[m].attrs
        if 'image_type' in attrs and attrs['image_type'] == 'Mask':
            masks_rec[m] = {
                'box': {
                    'blc': np.array([1., 1., 1., 1.]), 'comment': '', 'isRegion': 1,
                    'name': 'LCBox', 'oneRel': True, 'shape': np.array(casa_image_shape),
                    'trc': np.array(casa_image_shape)
                },
                'comment': '', 'isRegion': 1, 'name': 'LCPagedMask',
                'mask': f'Table: {os.sep.join([image_full_path, m])}'
            }
            masks.append(m)
    myvars = [sky_ap]
    myvars.extend(masks)
    for v in myvars:
        __write_pixels(v, active_mask, image_full_path, xds)
    if masks:
        tb = tables.table(
            image_full_path,
            readonly=False, lockoptions={'option': 'permanentwait'}, ack=False
        )
        tb.putkeyword('masks', masks_rec)
        tb.putkeyword('Image_defaultmask', active_mask)
        tb.close()


def __write_image_block(xda:xr.DataArray, outfile:str, blc:tuple) -> None:
    """
    Write image xda chunk to the corresponding image table slice
    """
    # trigger the DAG for this chunk and return values while the table is
    # unlocked
    values = xda.compute().values
    tb_tool = tables.table(
        outfile, readonly=False, lockoptions={'option': 'permanentwait'},
        ack=False
    )
    tb_tool.putcellslice(
        tb_tool.colnames()[0], 0, values, blc, tuple(
            np.array(blc) + np.array(values.shape) - 1
        )
    )
    tb_tool.close()


def __write_pixels(
    v: str, active_mask: str, image_full_path: str, xds: xr.Dataset
) -> None:
    flip = False
    if v == 'sky' or v == 'apeture':
        filename = image_full_path
    else:
        # mask
        flip = True
        filename = os.sep.join([image_full_path, v])
        if not os.path.exists(filename):
            tb = tables.table(os.sep.join([image_full_path, active_mask]))
            tb.copy(filename, deep=True, valuecopy=True)
            tb.close()
    arr = xds[v].isel(time=0).transpose(*('freq', 'pol', 'm', 'l'))
    chunk_bounds = arr.chunks
    b = [0, 0, 0, 0]
    loc0, loc1, loc2, loc3 = (0, 0, 0, 0)
    for i0 in chunk_bounds[0]:
        b[0] = loc0
        s0 = slice(b[0], b[0] + i0)
        loc1 = 0
        for i1 in chunk_bounds[1]:
            b[1] = loc1
            s1 = slice(b[1], b[1] + i1)
            loc2 = 0
            for i2 in chunk_bounds[2]:
                b[2] = loc2
                s2 = slice(b[2], b[2] + i2)
                loc3 = 0
                for i3 in chunk_bounds[3]:
                    b[3] = loc3
                    blc = tuple(b)
                    s3 = slice(b[3], b[3] + i3)
                    sub_arr = arr[s0, s1, s2, s3]
                    if flip:
                        sub_arr = np.logical_not(sub_arr)
                    __write_image_block(sub_arr, filename, blc)
                    loc3 += i3
                loc2 += i2
            loc1 += i1
        loc0 += i0





import pandas as pd


class _processing_set(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def summary(self):
        summary_data = {"name": [], "intent": [], "field_name": [], "frequency": []}
        for key, value in self.items():
            summary_data["name"].append(key)
            summary_data["intent"].append(value.attrs["intent"])
            summary_data["field_name"].append(value.attrs["field_info"]["name"])
            summary_data["frequency"].append(value["frequency"].values)
        summary_df = pd.DataFrame(summary_data)
        return summary_df



# from numcodecs.zstd import Zstd
import numcodecs
from typing import Dict, List, Tuple, Union
import itertools
import json
import numbers
from xradio.vis._vis_utils._ms.partitions import (
    finalize_partitions,
    read_ms_ddi_partitions,
    read_ms_scan_subscan_partitions,
    make_spw_names_by_ddi,
    make_partition_ids_by_ddi_intent,
    make_partition_ids_by_ddi_scan,
)

import dask
from xradio.vis._vis_utils._ms.descr import describe_ms
from xradio.vis._vis_utils._ms.msv2_msv3 import ignore_msv2_cols
from xradio.vis._vis_utils._ms._tables.read import (
    read_generic_table,
    make_freq_attrs,
    convert_casacore_time,
    extract_table_attributes,
)
from xradio.vis._vis_utils._ms._tables.read_main_table import (
    read_flat_main_table,
    read_expanded_main_table,
    get_baselines,
    get_utimes_tol,
    read_main_table_chunks,
)
from xradio.vis._vis_utils._ms.subtables import (
    subt_rename_ids,
    add_pointing_to_partition,
)
from xradio.vis._vis_utils._ms._tables.table_query import open_table_ro, open_query
from xradio.vis._vis_utils._utils.stokes_types import stokes_types
import numpy as np
from casacore import tables
from itertools import cycle
import logging
import time
import xarray as xr


def add_encoding(xds, compressor, chunks=None):
    if chunks is None:
        chunks = xds.dims

    chunks = {**dict(xds.dims), **chunks}  # Add missing dims if presents.

    encoding = {}
    for da_name in list(xds.data_vars):
        if chunks:
            da_chunks = [chunks[dim_name] for dim_name in xds[da_name].dims]
            xds[da_name].encoding = {"compressor": compressor, "chunks": da_chunks}
            # print(xds[da_name].encoding)
        else:
            xds[da_name].encoding = {"compressor": compressor}


def calc_indx_for_row_split(tb_tool, taql_where):
    baselines = get_baselines(tb_tool)
    col_names = tb_tool.colnames()
    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in col_names
        if tb_tool.iscelldefined(col, 0)
    ]

    freq_cnt, pol_cnt = [(cc[0], cc[1]) for cc in cshapes if len(cc) == 2][0]
    utimes, tol = get_utimes_tol(tb_tool, taql_where)
    # utimes = np.unique(tb_tool.getcol("TIME"))

    tvars = {}

    # chunks = [len(utimes), len(baselines), freq_cnt, pol_cnt]

    # print("nrows",  len(tb_tool.getcol("TIME")))

    tidxs = np.searchsorted(utimes, tb_tool.getcol("TIME"))

    ts_ant1, ts_ant2 = (
        tb_tool.getcol("ANTENNA1"),
        tb_tool.getcol("ANTENNA2"),
    )

    ts_bases = [
        str(ll[0]).zfill(3) + "_" + str(ll[1]).zfill(3)
        for ll in np.hstack([ts_ant1[:, None], ts_ant2[:, None]])
    ]
    bidxs = np.searchsorted(baselines, ts_bases)

    # some antenna 2"s will be out of bounds for this chunk, store rows that are in bounds
    didxs = np.where((bidxs >= 0) & (bidxs < len(baselines)))[0]

    baseline_ant1_id, baseline_ant2_id = np.array(
        [tuple(map(int, x.split("_"))) for x in baselines]
    ).T
    return (
        tidxs,
        bidxs,
        didxs,
        baseline_ant1_id,
        baseline_ant2_id,
        convert_casacore_time(utimes, False),
    )


def _check_single_field(tb_tool):
    field_id = np.unique(tb_tool.getcol("FIELD_ID"))
    # print(np.unique(field_id))
    assert len(field_id) == 1, "More than one field present."
    return field_id[0]


def _check_interval_consistent(tb_tool):
    interval = np.unique(tb_tool.getcol("INTERVAL"))
    assert len(interval) == 1, "Interval is not consistent."
    return interval[0]


def read_col(
    tb_tool,
    col: str,
    cshape: Tuple[int],
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
):
    start = time.time()
    data = tb_tool.getcol(col)
    # logging.info("Time to get col " + col + "  " + str(time.time()-start))

    # full data is the maximum of the data shape and chunk shape dimensions
    start = time.time()
    fulldata = np.full(cshape + data.shape[1:], np.nan, dtype=data.dtype)
    # logging.info("Time to full " + col + "  " + str(time.time()-start))

    start = time.time()
    fulldata[tidxs, bidxs] = data
    # logging.info("Time to reorganize " + col + "  " + str(time.time()-start))

    return fulldata


def create_attribute_metadata(col, tb_tool):
    # Still a lot to do.
    attrs_metadata = {}

    if col == "UVW":
        # Should not be hardcoded
        attrs_metadata["type"] = "uvw"
        attrs_metadata["units"] = "m"
        attrs_metadata["description"] = "uvw coordinates."

    return attrs_metadata


def create_coordinates(
    xds, infile, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id
):
    coords = {
        "time": utime,
        "baseline_antenna1_id": ("baseline_id", baseline_ant1_id),
        "baseline_antenna2_id": ("baseline_id", baseline_ant2_id),
        "uvw_label": ["u", "v", "w"],
        "baseline_id": np.arange(len(baseline_ant1_id)),
    }

    ddi_xds = read_generic_table(infile, "DATA_DESCRIPTION").sel(row=ddi)
    pol_setup_id = ddi_xds.polarization_id.values
    spw_id = ddi_xds.spectral_window_id.values

    spw_xds = read_generic_table(
        infile,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    ).sel(spectral_window_id=spw_id)
    coords["frequency"] = spw_xds["chan_freq"].data[
        ~(np.isnan(spw_xds["chan_freq"].data))
    ]

    pol_xds = read_generic_table(
        infile,
        "POLARIZATION",
        rename_ids=subt_rename_ids["POLARIZATION"],
    )
    num_corr = int(pol_xds["num_corr"][pol_setup_id].values)
    coords["polarization"] = np.vectorize(stokes_types.get)(
        pol_xds["corr_type"][pol_setup_id, :num_corr].values
    )

    xds = xds.assign_coords(coords)

    # Add metadata to coordinates:
    measures_freq_ref = spw_xds["meas_freq_ref"].data
    xds.frequency.attrs["type"] = "spectral_coord"
    xds.frequency.attrs["units"] = spw_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]["CHAN_FREQ"]["keywords"]["QuantumUnits"][0]
    xds.frequency.attrs["velocity_frame"] = spw_xds.attrs["other"]["msv2"][
        "ctds_attrs"
    ]["column_descriptions"]["CHAN_FREQ"]["keywords"]["MEASINFO"]["TabRefTypes"][
        measures_freq_ref
    ]
    xds.frequency.attrs["spectral_window_name"] = str(spw_xds.name.values)
    xds.frequency.attrs["reference_frequency"] = float(spw_xds.ref_frequency.values)
    xds.frequency.attrs["effective_channel_width"] = "EFFECTIVE_CHANNEL_WIDTH"
    # Add if doppler table is present
    # xds.frequency.attrs["doppler_velocity"] =
    # xds.frequency.attrs["doppler_type"] =

    # unique_chan_width = np.unique(spw_xds.chan_width.data[np.logical_not(np.isnan(spw_xds.chan_width.data))])
    # print('unique_chan_width',unique_chan_width)
    # print('spw_xds.chan_width.data',spw_xds.chan_width.data)
    # assert len(unique_chan_width) == 1, "Channel width varies for spw."
    xds.frequency.attrs["channel_width"] = spw_xds.chan_width.data[
        ~(np.isnan(spw_xds.chan_width.data))
    ]  # unique_chan_width[0]

    main_table_attrs = extract_table_attributes(infile)
    xds.time.attrs["type"] = "time"
    xds.time.attrs["units"] = main_table_attrs["column_descriptions"]["TIME"][
        "keywords"
    ]["QuantumUnits"][0]
    xds.time.attrs["time_scale"] = main_table_attrs["column_descriptions"]["TIME"][
        "keywords"
    ]["MEASINFO"]["Ref"]
    xds.time.attrs[
        "format"
    ] = "unix"  # Time gets converted to unix in xradio.vis._vis_utils._ms._tables.read.convert_casacore_time
    xds.time.attrs["integration_time"] = interval
    xds.time.attrs["effective_integration_time"] = "EFFECTIVE_INTEGRATION_TIME"

    return xds


def convert_and_write_partition(
    infile: str,
    outfile: str,
    intent: str,
    ddi: int = 0,
    state_ids=None,
    field_id: int = None,
    ignore_msv2_cols: Union[list, None] = None,
    chunks_on_disk: Union[Dict, None] = None,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    storage_backend="zarr",
    overwrite: bool = False,
):
    if ignore_msv2_cols is None:
        ignore_msv2_cols = []

    file_name = (
        outfile
        + "/"
        + outfile.replace(".vis.zarr", "").split("/")[-1]
        + "_ddi_"
        + str(ddi)
        + "_intent_"
        + intent
    )
    taql_where = f"where (DATA_DESC_ID = {ddi})"

    if isinstance(state_ids, numbers.Integral):
        taql_where += f" AND (STATE_ID = {state_ids})"
    else:
        state_ids_or = " OR STATE_ID = ".join(np.char.mod("%d", state_ids))
        taql_where += f" AND (STATE_ID = {state_ids_or})"

    if field_id is not None:
        taql_where += f" AND (FIELD_ID = {field_id})"
        file_name = file_name + "_field_id_" + str(field_id)

    ddi_xds = read_generic_table(infile, "DATA_DESCRIPTION").sel(row=ddi)
    spw_id = ddi_xds.spectral_window_id.values

    spw_xds = read_generic_table(
        infile,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    ).sel(spectral_window_id=spw_id)
    n_chan = len(spw_xds["chan_freq"].data[~(np.isnan(spw_xds["chan_freq"].data))])

    start_with = time.time()
    with open_table_ro(infile) as mtable:
        # one partition, select just the specified ddi (+ scan/subscan)
        taql_main = f"select * from $mtable {taql_where}"
        with open_query(mtable, taql_main) as tb_tool:
            # print(taql_where,file_name,"Flag shape",tb_tool.getcol('FLAG').shape)

            if tb_tool.nrows() == 0:
                tb_tool.close()
                mtable.close()
                return xr.Dataset(), {}, {}

            # logging.info("Setting up table "+ str(time.time()-start_with))

            start = time.time()
            (
                tidxs,
                bidxs,
                didxs,
                baseline_ant1_id,
                baseline_ant2_id,
                utime,
            ) = calc_indx_for_row_split(tb_tool, taql_where)
            time_baseline_shape = (len(utime), len(baseline_ant1_id))
            # logging.info("Calc indx for row split "+ str(time.time()-start))

            start = time.time()
            xds = xr.Dataset()
            col_to_data_variable_names = {
                "FLOAT_DATA" : "SPECTRUM",
                "DATA": "VISIBILITY",
                "CORRECTED_DATA": "VISIBILITY_CORRECTED",
                "WEIGHT_SPECTRUM": "WEIGHT",
                "WEIGHT": "WEIGHT",
                "FLAG": "FLAG",
                "UVW": "UVW",
                "TIME_CENTROID": "TIME_CENTROID",
                "EXPOSURE": "EFFECTIVE_INTEGRATION_TIME",
            }
            col_dims = {
                "DATA": ("time", "baseline_id", "frequency", "polarization"),
                "CORRECTED_DATA": ("time", "baseline_id", "frequency", "polarization"),
                "WEIGHT_SPECTRUM": ("time", "baseline_id", "frequency", "polarization"),
                "WEIGHT": ("time", "baseline_id", "frequency", "polarization"),
                "FLAG": ("time", "baseline_id", "frequency", "polarization"),
                "UVW": ("time", "baseline_id", "uvw_label"),
                "TIME_CENTROID": ("time", "baseline_id"),
                "EXPOSURE": ("time", "baseline_id"),
                "FLOAT_DATA": ("time", "baseline_id", "frequency", "polarization"),
            }
            col_to_coord_names = {
                "TIME": "time",
                "ANTENNA1": "baseline_ant1_id",
                "ANTENNA2": "baseline_ant2_id",
            }
            coords_dim_select = {
                "TIME": np.s_[:, 0:1],
                "ANTENNA1": np.s_[0:1, :],
                "ANTENNA2": np.s_[0:1, :],
            }
            check_variables = {}

            col_names = tb_tool.colnames()

            # Create Data Variables
            # logging.info("Setup xds "+ str(time.time()-start))
            for col in col_names:
                if col in col_to_data_variable_names:
                    if (col == "WEIGHT") and ("WEIGHT_SPECTRUM" in col_names):
                        continue
                    try:
                        start = time.time()
                        if col == "WEIGHT":
                            xds[col_to_data_variable_names[col]] = xr.DataArray(
                                np.tile(
                                    read_col(
                                        tb_tool,
                                        col,
                                        time_baseline_shape,
                                        tidxs,
                                        bidxs,
                                        didxs,
                                    )[:, :, None, :],
                                    (1, 1, n_chan, 1),
                                ),
                                dims=col_dims[col],
                            )

                        else:
                            xds[col_to_data_variable_names[col]] = xr.DataArray(
                                read_col(
                                    tb_tool,
                                    col,
                                    time_baseline_shape,
                                    tidxs,
                                    bidxs,
                                    didxs,
                                ),
                                dims=col_dims[col],
                            )
                            # logging.info("Time to read column " + str(col) + " : " + str(time.time()-start))
                    except:
                        # logging.debug("Could not load column",col)
                        continue

                    xds[col_to_data_variable_names[col]].attrs.update(
                        create_attribute_metadata(col, tb_tool)
                    )

            field_id = _check_single_field(tb_tool)
            interval = _check_interval_consistent(tb_tool)

            start = time.time()

            xds = create_coordinates(
                xds, infile, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id
            )

            field_xds = read_generic_table(
                infile,
                "FIELD",
                rename_ids=subt_rename_ids["FIELD"],
            )

            field_info = {
                "name": field_xds["name"].data[field_id],
                "code": field_xds["code"].data[field_id],
                "time": field_xds["time"].data[field_id],
                "num_poly": 0,
                "delay_dir": list(field_xds["delay_dir"].data[field_id, 0, :]),
                "phase_dir": list(field_xds["phase_dir"].data[field_id, 0, :]),
                "reference_dir": list(field_xds["reference_dir"].data[field_id, 0, :]),
            }
            xds.attrs["field_info"] = field_info

            xds.attrs["data_groups"] = {
                "base": {
                    "visibility": "VISIBILITY",
                    "flag": "FLAG",
                    "weight": "WEIGHT",
                    "uvw": "UVW",
                }
            }

            if overwrite:
                mode = "w"
            else:
                mode = "w-"

            add_encoding(xds, compressor=compressor, chunks=chunks_on_disk)

            ant_xds = read_generic_table(
                infile,
                "ANTENNA",
                rename_ids=subt_rename_ids["ANTENNA"],
            )
            del ant_xds.attrs["other"]

            xds.attrs["intent"] = intent
            
            #Time and frequency should always be increasing
            if xds.frequency[1]-xds.frequency[0] < 0:
                xds = xds.sel(frequency=slice(None, None, -1))
            
            if xds.time[1]-xds.time[0] < 0:
                xds = xds.sel(time=slice(None, None, -1))

            if storage_backend == "zarr":
                xds.to_zarr(store=file_name + "/MAIN", mode=mode)
                ant_xds.to_zarr(store=file_name + "/ANTENNA", mode=mode)
            elif storage_backend == "netcdf":
                # xds.to_netcdf(path=file_name+"/MAIN", mode=mode) #Does not work
                raise

    # logging.info("Saved ms_v4 " + file_name + " in " + str(time.time() - start_with) + "s")


def get_unqiue_intents(infile):
    state_xds = read_generic_table(
        infile,
        "STATE",
        rename_ids=subt_rename_ids["STATE"],
    )

    obs_mode_dict = {}
    for i, obs_mode in enumerate(state_xds.obs_mode.values):
        if obs_mode in obs_mode_dict:
            obs_mode_dict[obs_mode].append(i)
        else:
            obs_mode_dict[obs_mode] = [i]

    return list(obs_mode_dict.keys()), obs_mode_dict.values()


def enumerated_product(*args):
    yield from zip(
        itertools.product(*(range(len(x)) for x in args)), itertools.product(*args)
    )


def convert_msv2_to_processing_set(
    infile: str,
    outfile: str,
    partition_scheme: str,  # intent_field, subscan
    chunks_on_disk: Union[Dict, None] = None,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    parallel: bool = False,
    storage_backend="zarr",
    overwrite: bool = False,
):
    """ """
    spw_xds = read_generic_table(
        infile,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    )

    ddi_xds = read_generic_table(infile, "DATA_DESCRIPTION")
    data_desc_ids = np.arange(ddi_xds.dims["row"])

    if partition_scheme == "ddi_intent_field":
        unique_intents, state_ids = get_unqiue_intents(infile)
        field_ids = np.arange(read_generic_table(infile, "FIELD").dims["row"])
    elif partition_scheme == "ddi_state":
        state_xds = read_generic_table(infile, "STATE")
        state_ids = np.arange(state_xds.dims["row"])
        intents = state_xds.obs_mode.values
        # print(state_xds, intents)
        # field_ids = [None]
        field_ids = np.arange(read_generic_table(infile, "FIELD").dims["row"])

    delayed_list = []
    partitions = {}
    cnt = 0

    # for ddi, state, field in itertools.product(data_desc_ids, state_ids, field_ids):
    #    logging.info("DDI " + str(ddi) + ", STATE " + str(state) + ", FIELD " + str(field))

    for idx, pair in enumerated_product(data_desc_ids, state_ids, field_ids):
        ddi, state_id, field_id = pair
        # logging.debug("DDI " + str(ddi) + ", STATE " + str(state_id) + ", FIELD " + str(field_id))

        if partition_scheme == "ddi_intent_field":
            intent = unique_intents[idx[1]]
        else:
            intent = intents[idx[1]] + "_" + str(state_id)

        if parallel:
            delayed_list.append(
                dask.delayed(convert_and_write_partition)(
                    infile,
                    outfile,
                    intent,
                    ddi,
                    state_id,
                    field_id,
                    ignore_msv2_cols=ignore_msv2_cols,
                    chunks_on_disk=chunks_on_disk,
                    compressor=compressor,
                    overwrite=overwrite,
                )
            )
        else:
            convert_and_write_partition(
                infile,
                outfile,
                intent,
                ddi,
                state_id,
                field_id,
                ignore_msv2_cols=ignore_msv2_cols,
                chunks_on_disk=chunks_on_disk,
                compressor=compressor,
                storage_backend=storage_backend,
                overwrite=overwrite,
            )

    if parallel:
        dask.compute(delayed_list)



import os
import xradio

def download(file, folder='.', unpack=False, source='dropbox'):

  if os.path.exists(folder) == False:
    os.mkdir(folder)
  
  if source == 'gdrive':
    from xradio.data._google_drive import download
    download(file=file, folder=folder, unpack=unpack)

  elif source == 'dropbox':
    from xradio.data._dropbox import download
    download(file=file, folder=folder)

  else:
    print("unknown source or issue found")
    pass



import logging
import os
from typing import Union
import xarray as xr
from ._fits.fits_to_xds import __fits_image_to_xds_metadata

def __read_fits_image(
    infile:str, chunks:Union[list, dict], masks:bool=True,
    verbose:bool=False
) -> xr.Dataset:
    img_full_path = os.path.expanduser(infile)
    attrs = __fits_image_to_xds_metadata(img_full_path, verbose)
    # return attrs
    """
    xds = ret['xds']
    dimorder = __get_xds_dim_order(ret['sphr_dims'])
    xds = __add_sky_or_apeture(
        xds, __read_image_array(img_full_path, chunks, verbose=verbose),
        dimorder, img_full_path, ret['sphr_dims']
    )
    if masks:
        mymasks = __get_mask_names(img_full_path)
        for m in mymasks:
            ary = __read_image_array(img_full_path, chunks, mask=m, verbose=verbose)
            xds = __add_mask(xds, m, ary, dimorder)
    xds.attrs = __casa_image_to_xds_attrs(img_full_path, history)
    mb = __multibeam_array(xds, img_full_path)
    if mb is not None:
        xds['beam'] = mb
    xds = __add_coord_attrs(xds, ret['icoords'], ret['dir_axes'])
    return xds
    """





import astropy.units as u
import dask
import dask.array as da
import xarray as xr


__c = 2.99792458e+08 * u.m/u.s


def __get_xds_dim_order(has_sph:bool) -> list:
    dimorder = ['time', 'pol', 'freq']
    dir_lin = ['l', 'm'] if has_sph else ['u', 'v']
    dimorder.extend(dir_lin)
    return dimorder


def __coords_to_numpy(xds):
    for k, v in xds.coords.items():
        if dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.assign_coords({k: (v.dims, v.to_numpy())})
            xds[k].attrs = attrs
    return xds


def __dask_arrayize(xds):
    """
    If necessary, change coordinates to numpy arrays and data
    variables to dask arrays
    """
    xds = __coords_to_numpy(xds)
    for k, v in xds.data_vars.items():
        if not dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.drop_vars([k])
            # may need to add dims to this call as in numpy method analogs in this file
            xds = xds.assign({k: da.array(v)})
            xds[k].attrs = attrs
    for k, v in xds.attrs.items():
        if isinstance(v, xr.Dataset):
            xds.attrs[k] = __dask_arrayize(v)
    return xds


def __numpy_arrayize(xds):
    xds = __coords_to_numpy(xds)
    for k, v in xds.data_vars.items():
        if dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.drop_vars([k])
            xds = xds.assign({k: (v.dims, v.to_numpy())})
            xds[k].attrs = attrs
    for k, v in xds.attrs.items():
        if isinstance(v, xr.Dataset):
            xds.attrs[k] = __dask_arrayize(v)
    return xds


